#cython: language_level=3

cdef csleqp.SLEQP_RETCODE sleqp_lsq_nonzeros(csleqp.SleqpFunc* func,
                                             int* residual_nnz,
                                             int* jac_fwd_nnz,
                                             int* jac_adj_nnz,
                                             int* cons_val_nnz,
                                             int* cons_jac_nnz,
                                             void* func_data):

  try:
    func_obj = (<object> func_data)

    def try_call(f):
      try:
        return f()
      except AttributeError:
        return csleqp.SLEQP_NONE

    residual_nnz[0] = try_call(lambda : func_obj.residual_nnz())
    jac_fwd_nnz[0] = try_call(lambda : func_obj.jac_fwd_nnz())
    jac_adj_nnz[0] = try_call(lambda : func_obj.jac_adj_nnz())
    cons_val_nnz[0] = try_call(lambda : func_obj.cons_val_nnz())
    cons_jac_nnz[0] = try_call(lambda : func_obj.cons_jac_nnz())

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_lsq_nonzeros_nogil(csleqp.SleqpFunc* func,
                                                   int* residual_nnz,
                                                   int* jac_fwd_nnz,
                                                   int* jac_adj_nnz,
                                                   int* cons_val_nnz,
                                                   int* cons_jac_nnz,
                                                   void* func_data) nogil:

  with gil:
    return sleqp_lsq_nonzeros(func,
                              residual_nnz,
                              jac_fwd_nnz,
                              jac_adj_nnz,
                              cons_val_nnz,
                              cons_jac_nnz,
                              func_data)


cdef csleqp.SLEQP_RETCODE sleqp_lsq_residuals(csleqp.SleqpFunc* func,
                                              csleqp.SleqpVec* residual,
                                              void* func_data):
  try:
    func_obj = (<object> func_data)

    array = func_obj.lsq_residuals()
    csleqp_call(array_to_sleqp_sparse_vec(array, residual))

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_lsq_residuals_nogil(csleqp.SleqpFunc* func,
                                                    csleqp.SleqpVec* residual,
                                                    void* func_data) nogil:
  with gil:
    return sleqp_lsq_residuals(func,
                               residual,
                               func_data)


cdef csleqp.SLEQP_RETCODE sleqp_lsq_jac_forward(csleqp.SleqpFunc* func,
                                                const csleqp.SleqpVec* forward_direction,
                                                csleqp.SleqpVec* product,
                                                void* func_data):
  try:
      func_obj = (<object> func_data)

      forward_direction_array = _readonly(sleqp_sparse_vec_to_array(forward_direction))

      product_array = func_obj.lsq_jac_forward(forward_direction_array)

      assert product_array is not None, "lsq_jac_forward(...) returned 'None'"

      csleqp_call(array_to_sleqp_sparse_vec(product_array, product))

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_lsq_jac_forward_nogil(csleqp.SleqpFunc* func,
                                                      const csleqp.SleqpVec* forward_direction,
                                                      csleqp.SleqpVec* product,
                                                      void* func_data) nogil:
  with gil:
    return sleqp_lsq_jac_forward(func,
                                 forward_direction,
                                 product,
                                 func_data)


cdef csleqp.SLEQP_RETCODE sleqp_lsq_jac_adjoint(csleqp.SleqpFunc* func,
                                                const csleqp.SleqpVec* adjoint_direction,
                                                csleqp.SleqpVec* product,
                                                void* func_data):
  try:
      func_obj = (<object> func_data)

      adjoint_direction_array = _readonly(sleqp_sparse_vec_to_array(adjoint_direction))

      product_array = func_obj.lsq_jac_adjoint(adjoint_direction_array)

      assert product_array is not None, "lsq_jac_adjoint(...) returned 'None'"

      csleqp_call(array_to_sleqp_sparse_vec(product_array, product))
  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_lsq_jac_adjoint_nogil(csleqp.SleqpFunc* func,
                                                      const csleqp.SleqpVec* adjoint_direction,
                                                      csleqp.SleqpVec* product,
                                                      void* func_data) nogil:
  with gil:
    return sleqp_lsq_jac_adjoint(func,
                                 adjoint_direction,
                                 product,
                                 func_data)


cdef set_lsq_func_callbacks(csleqp.SleqpLSQCallbacks* callbacks):
  if release_gil:
    callbacks.set_value            = &sleqp_func_set_nogil
    callbacks.lsq_nonzeros         = &sleqp_lsq_nonzeros_nogil
    callbacks.lsq_residuals        = &sleqp_lsq_residuals_nogil
    callbacks.lsq_jac_forward      = &sleqp_lsq_jac_forward_nogil
    callbacks.lsq_jac_adjoint      = &sleqp_lsq_jac_adjoint_nogil
    callbacks.cons_val             = &sleqp_func_cons_val_nogil
    callbacks.cons_jac             = &sleqp_func_cons_jac_nogil
  else:
    callbacks.set_value            = &sleqp_func_set
    callbacks.lsq_nonzeros         = &sleqp_lsq_nonzeros
    callbacks.lsq_residuals        = &sleqp_lsq_residuals
    callbacks.lsq_jac_forward      = &sleqp_lsq_jac_forward
    callbacks.lsq_jac_adjoint      = &sleqp_lsq_jac_adjoint
    callbacks.cons_val             = &sleqp_func_cons_val
    callbacks.cons_jac             = &sleqp_func_cons_jac

  callbacks.func_free = &sleqp_func_free


cdef object lsq_funcs = weakref.WeakSet()

cdef update_lsq_func_callbacks():
  cdef _Func func
  cdef csleqp.SleqpLSQCallbacks callbacks

  set_lsq_func_callbacks(&callbacks)

  for obj in lsq_funcs:
    func = <_Func> obj

    csleqp_call(csleqp.sleqp_lsq_func_set_callbacks(func.cfunc,
                                                    &callbacks))

cdef csleqp.SLEQP_RETCODE create_lsq_func(csleqp.SleqpFunc** cfunc,
                                          object func,
                                          int num_variables,
                                          int num_constraints,
                                          int num_residuals,
                                          double levenberg_marquardt,
                                          csleqp.SleqpSettings* settings):
  cdef csleqp.SleqpLSQCallbacks callbacks

  assert func is not None

  set_lsq_func_callbacks(&callbacks)

  return csleqp.sleqp_lsq_func_create(cfunc,
                                      &callbacks,
                                      num_variables,
                                      num_constraints,
                                      num_residuals,
                                      levenberg_marquardt,
                                      settings,
                                      <void*> func)



#cython: language_level=3

from enum import Enum, Flag, auto


class _DocEnum(Enum):
    def __new__(cls, value, doc=None):
        self = object.__new__(cls)  # calling super().__new__(value) here would fail
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
            self.desc    = doc
        return self


class Status(_DocEnum):
  """
  The status resulting from the solution process
  """
  Unkown         = csleqp.SLEQP_STATUS_UNKNOWN, "Unknown state"
  Running        = csleqp.SLEQP_STATUS_RUNNING, "Solver is running"
  Optimal        = csleqp.SLEQP_STATUS_OPTIMAL, "An optimal solution was found"
  Infeasible     = csleqp.SLEQP_STATUS_INFEASIBLE, "Problem was detected to be infeasible"
  Unbounded      = csleqp.SLEQP_STATUS_UNBOUNDED, "Problem appears unbounded"
  AbortDeadpoint = csleqp.SLEQP_STATUS_ABORT_DEADPOINT, "Reached a dead point"
  AbortIter      = csleqp.SLEQP_STATUS_ABORT_ITER, "Aborted after reaching iteration limit"
  AbortManual    = csleqp.SLEQP_STATUS_ABORT_MANUAL, "Manual abort requested"
  AbortTime      = csleqp.SLEQP_STATUS_ABORT_TIME, "Aborted after reaching time limit"


class DerivCheck(Flag):
  Skip             = csleqp.SLEQP_DERIV_CHECK_SKIP
  FirstFunc        = csleqp.SLEQP_DERIV_CHECK_FIRST_OBJ
  FirstCons        = csleqp.SLEQP_DERIV_CHECK_FIRST_CONS
  First            = csleqp.SLEQP_DERIV_CHECK_FIRST
  SecondFunc       = csleqp.SLEQP_DERIV_CHECK_SECOND_OBJ
  SecondCons       = csleqp.SLEQP_DERIV_CHECK_SECOND_CONS
  SecondExhaustive = csleqp.SLEQP_DERIV_CHECK_SECOND_EXHAUSTIVE
  SecondSimple     = csleqp.SLEQP_DERIV_CHECK_SECOND_SIMPLE


class HessianEval(_DocEnum):
  """
  The evaluation method used for Hessian products
  """
  Exact      = csleqp.SLEQP_HESS_EVAL_EXACT, "Exact evaluation"
  SR1        = csleqp.SLEQP_HESS_EVAL_SR1, "The SR1 quasi-Newton method"
  SimpleBFGS = csleqp.SLEQP_HESS_EVAL_SIMPLE_BFGS, "The BFGS method for convex functions"
  DampedBFGS = csleqp.SLEQP_HESS_EVAL_DAMPED_BFGS, "A damped BFGS method "


class Sizing(Enum):
  NoSizing   = csleqp.SLEQP_BFGS_SIZING_NONE
  CenteredOL = csleqp.SLEQP_BFGS_SIZING_CENTERED_OL


class TRSolver(_DocEnum):
  """
  The trust-region solver used
  """
  TRlib = csleqp.SLEQP_TR_SOLVER_TRLIB, "The trlib implementation of the Generalized Lanczos method"
  CG    = csleqp.SLEQP_TR_SOLVER_CG, "Steihaug's conjugate gradient method"
  LSQR  = csleqp.SLEQP_TR_SOLVER_LSQR, "LSQR solver for LSQ functions"
  Auto  = csleqp.SLEQP_TR_SOLVER_AUTO, "Automatically chosen"


class PolishingType(_DocEnum):
  """
  The polishing methods use
  """
  NoPolishing = csleqp.SLEQP_POLISHING_NONE, "No polishing"
  ZeroDual    = csleqp.SLEQP_POLISHING_ZERO_DUAL, "Remove bounds and constraints with zero duals"
  Inactive    = csleqp.SLEQP_POLISHING_INACTIVE, "Remove inactive bounds and constraints"


class StepRule(_DocEnum):
  """
  The step rule used
  """
  Direct = csleqp.SLEQP_STEP_RULE_DIRECT, "Direct"
  Window = csleqp.SLEQP_STEP_RULE_WINDOW, "Sliding window"
  MinStep = csleqp.SLEQP_STEP_RULE_MINSTEP, "Min-step"


class LineSearch(_DocEnum):
  """
  The linesearch used
  """
  Exact  = csleqp.SLEQP_LINESEARCH_EXACT, "Exact line search"
  Approx = csleqp.SLEQP_LINESEARCH_APPROX, "Approximate line search"


class ParametricCauchy(_DocEnum):
  """
  The parametric Cauchy type to be used
  """
  Disabled = csleqp.SLEQP_PARAMETRIC_CAUCHY_DISABLED, "Disable parametric Cauchy"
  Coarse   = csleqp.SLEQP_PARAMETRIC_CAUCHY_COARSE, "Coarse parametric Cauchy"
  Fine     = csleqp.SLEQP_PARAMETRIC_CAUCHY_FINE, "Fine parametric Cauchy"


class InitialTRChoice(_DocEnum):
  """
  Choice for initial trust radius
  """
  Narrow = csleqp.SLEQP_INITIAL_TR_CHOICE_NARROW, "Narrow"
  Wide   = csleqp.SLEQP_INITIAL_TR_CHOICE_WIDE, "Wide"


class AugJacMethod(_DocEnum):
  """
  Augmented Jacobian type
  """
  Auto = csleqp.SLEQP_AUG_JAC_AUTO, "Auto"
  Standard = csleqp.SLEQP_AUG_JAC_STANDARD, "Standard"
  Reduced = csleqp.SLEQP_AUG_JAC_REDUCED, "Reduced"
  Direct = csleqp.SLEQP_AUG_JAC_DIRECT, "Direct"


class ValueReason(_DocEnum):
  """
  The reason for setting a new function value
  """
  NoReason         = csleqp.SLEQP_VALUE_REASON_NONE, "No reason given"
  Init             = csleqp.SLEQP_VALUE_REASON_INIT, "Initialization"
  CheckingDeriv    = csleqp.SLEQP_VALUE_REASON_CHECKING_DERIV, "Derivative check"
  AcceptedIterate  = csleqp.SLEQP_VALUE_REASON_ACCEPTED_ITERATE, "Accepted a trial iterate"
  TryingIterate    = csleqp.SLEQP_VALUE_REASON_TRYING_ITERATE, "Trying out a trial iterate"
  TryingSOCIterate = csleqp.SLEQP_VALUE_REASON_TRYING_SOC_ITERATE, "Trying a second-order correction"
  RejectedIterate  = csleqp.SLEQP_VALUE_REASON_REJECTED_ITERATE, "Rejected an iterate restoring the previous"


class ActiveState(_DocEnum):
  """
  The state of a variable or constraint in the working set
  """
  Inactive    = csleqp.SLEQP_INACTIVE, "Variable or constraint is inactive"
  ActiveLower = csleqp.SLEQP_ACTIVE_LOWER, "Variable or constraint is active at its lower bound"
  ActiveUpper = csleqp.SLEQP_ACTIVE_UPPER, "Variable or constraint is active at its upper bound"
  ActiveBoth  = csleqp.SLEQP_ACTIVE_BOTH, "Variable or constraint is active at both bounds (which must coincide)"


class DualEstimationType(_DocEnum):
  """
  The type of dual estimation used
  """
  LP      = csleqp.SLEQP_DUAL_ESTIMATION_TYPE_LP, "Dual variables of the LP"
  LSQ     = csleqp.SLEQP_DUAL_ESTIMATION_TYPE_LSQ, "Least-squares estimation"


class SolverEvent(Enum):
  AcceptedIterate    = csleqp.SLEQP_SOLVER_EVENT_ACCEPTED_ITERATE
  PerformedIteration = csleqp.SLEQP_SOLVER_EVENT_PERFORMED_ITERATION
  Finished           = csleqp.SLEQP_SOLVER_EVENT_FINISHED


class SolverState(Enum):
  TrustRadius              = auto()
  LPTrustRadius            = auto()
  ScaledFuncVal            = auto()
  ScaledMeritVal           = auto()
  ScaledFeasRes            = auto()
  ScaledStatRes            = auto()
  ScaledSlackRes           = auto()
  PenaltyParameter         = auto()
  MinRayleigh              = auto()
  MaxRayleigh              = auto()
  LastStepOnBoundary       = auto()
  Iteration                = auto()
  LastStepType             = auto()
  ScaledStatResiduals      = auto()
  ScaledFeasResiduals      = auto()
  ScaledConsSlackResiduals = auto()
  ScaledVarSlackResiduals  = auto()


class StepType(Enum):
  NoStep       = csleqp.SLEQP_STEPTYPE_NONE
  Accepted     = csleqp.SLEQP_STEPTYPE_ACCEPTED
  AcceptedFull = csleqp.SLEQP_STEPTYPE_ACCEPTED_FULL
  AcceptedSOC  = csleqp.SLEQP_STEPTYPE_ACCEPTED_SOC
  Rejected     = csleqp.SLEQP_STEPTYPE_REJECTED



#cython: language_level=3

from collections import defaultdict

class CallbackError(Exception):
  def __init__(self, *args, **kwds):
    super().__init__(*args, **kwds)

class EvaluationError(Exception):
  def __init__(self, *args, **kwds):
    super().__init__(*args, **kwds)

cdef object _exception_map = {
  csleqp.SLEQP_FAILED_ASSERTION: AssertionError,
  csleqp.SLEQP_NOMEM: MemoryError,
  csleqp.SLEQP_INTERNAL_ERROR: Exception,
  csleqp.SLEQP_FUNC_EVAL_ERROR: EvaluationError,
  csleqp.SLEQP_CALLBACK_ERROR: CallbackError,
  csleqp.SLEQP_MATH_ERROR: ArithmeticError,
  csleqp.SLEQP_INVALID_DERIV: Exception,
  csleqp.SLEQP_ILLEGAL_ARGUMENT: ValueError
}

cdef _get_exception():
  cdef csleqp.SLEQP_ERROR_TYPE error_type = csleqp.sleqp_error_type()
  cdef const char* error_msg = csleqp.sleqp_error_msg()
  msg = error_msg.decode('UTF-8')
  if error_type in _exception_map:
    exception = _exception_map[error_type]
    return exception(msg)
  return Exception(error_msg.decode('UTF-8'))

cdef _raise_exception():
  raise _get_exception()

cdef csleqp_call(csleqp.SLEQP_RETCODE retcode):
  if retcode != csleqp.SLEQP_OKAY:
    _raise_exception()



#cython: language_level=3

cdef object sleqp_sparse_vec_to_array(const csleqp.SleqpVec* vec):
  assert vec
  values = np.zeros((vec.dim,), dtype=np.float64)

  for k in range(vec.nnz):
    values[vec.indices[k]] = vec.data[k]

  return values

cdef csleqp.SLEQP_RETCODE array_to_sleqp_sparse_vec(np.ndarray array,
                                                    csleqp.SleqpVec* vec) except csleqp.SLEQP_ERROR:
  assert vec != NULL

  csleqp_call(csleqp.sleqp_vec_clear(vec))

  if array is None:
    return csleqp.SLEQP_OKAY

  assert array.ndim == 1

  cdef int dim = array.shape[0]

  csleqp_call(csleqp.sleqp_vec_reserve(vec, dim))

  for i in range(dim):
    if array[i] != 0.:
      csleqp_call(csleqp.sleqp_vec_push(vec,
                                                  i,
                                                  array[i]))

  return csleqp.SLEQP_OKAY

class DenseIterator:
  def __init__(self, matrix):
    self.matrix = matrix
    self.iter = np.nditer(self.matrix,
                          flags=['multi_index'],
                          order='F')

  def __iter__(self):
    while not self.iter.finished:
      data = self.iter[0]
      if not data:
        self.iter.iternext()
        continue

      yield self.iter.multi_index, data

      self.iter.iternext()

  def length_bound(self):
    return np.prod(self.matrix.shape)


class CSCIterator:
  def __init__(self, matrix):
    assert scipy.sparse.isspmatrix_csc(matrix)
    self.matrix = matrix
    self._length_bound = -1

  def __iter__(self):
    for col in range(self.matrix.shape[1]):
      for ind in range(self.matrix.indptr[col], self.matrix.indptr[col+1]):
        data = self.matrix.data[ind]
        if data:
          yield (self.matrix.indices[ind], col), self.matrix.data[ind]

  def length_bound(self):
    if self._length_bound != -1:
      return self._length_bound

    self._length_bound = self.matrix.count_nonzero()

    return self._length_bound

def matrix_iterator(matrix):
  assert not (matrix is None)
  assert matrix.ndim == 2

  if not scipy.sparse.issparse(matrix):
    return DenseIterator(matrix)

  assert scipy.sparse.issparse(matrix)

  if not scipy.sparse.isspmatrix_csc(matrix):
    matrix = scipy.sparse.csc_matrix(matrix)

  assert scipy.sparse.isspmatrix_csc(matrix)

  return CSCIterator(matrix)

cdef csleqp.SLEQP_RETCODE matrix_to_sleqp_sparse_matrix(object mat,
                                                        csleqp.SleqpMat* matrix) \
                                                        except csleqp.SLEQP_ERROR:
  assert matrix

  if mat is None:
    return csleqp.SLEQP_OKAY

  assert mat.ndim == 2

  num_rows = csleqp.sleqp_mat_num_rows(matrix)
  num_cols = csleqp.sleqp_mat_num_cols(matrix)

  assert mat.shape == (num_rows, num_cols)

  if num_rows == 0 or num_cols == 0:
    return csleqp.SLEQP_OKAY

  cdef int last_col = -1

  matrix_iter = matrix_iterator(mat)

  for (row, col), data in matrix_iter:
    while last_col != col:
      last_col += 1
      csleqp_call(csleqp.sleqp_mat_push_col(matrix, last_col))

    nnz = csleqp.sleqp_mat_nnz(matrix)
    nnz_max = csleqp.sleqp_mat_nnz_max(matrix)

    if nnz == nnz_max:
      csleqp_call(csleqp.sleqp_mat_reserve(matrix,
                                           matrix_iter.length_bound()))

    csleqp_call(csleqp.sleqp_mat_push(matrix,
                                      row,
                                      col,
                                      data))

  while last_col < csleqp.sleqp_mat_num_cols(matrix) - 1:
    last_col += 1
    csleqp_call(csleqp.sleqp_mat_push_col(matrix, last_col))

  return csleqp.SLEQP_OKAY



import numpy as np
import scipy.sparse as sps

from scipy.sparse.linalg import LinearOperator

from sleqp._derivative import FD_METHODS, create_derivative


class Hessian:
  def set_value(self, x):
    pass

  def product(self, args, grad, d):
    pass


class HessianProduct(Hessian):
  def __init__(self, hessp):
    self.hessp = hessp

  def set_value(self, x):
    self.x = x

  def product(self, args, direction, grad=None):
    return self.hessp(self.x, direction, *args)


# Callable Hessian
class SparseHessian(Hessian):
  def __init__(self, x, hess, hess_val):
    self.hess = hess
    self.x = x
    self.hval = hess_val

  def set_value(self, x):
    self.x = x
    self.hval = None

  def _eval(self, args):
    if self.hval is None:
      self.hval = self.hess(self.x, *args)

  def product(self, args, direction, grad=None):
    self._eval(args)
    return self.hval.dot(direction)


# Callable Hessian
class DenseHessian(Hessian):
  def __init__(self, x, hess, hess_val):
    self.hess = hess
    self.x = x
    self.hval = self._prepare_hess_val(hess_val)

  def set_value(self, x):
    self.x = x
    self.hval = None

  def _prepare_hess_val(self, hess_val):
    return np.atleast_2d(hess_val)

  def _eval(self, args):
    if self.hval is None:
      self.hval = self._prepare_hess_val(self.hess(self.x, *args))

  def product(self, args, direction, grad=None):
    self._eval(args)
    return np.dot(self.hval, direction)


# Callable Hessian
class HessianOperator(Hessian):
  def __init__(self, x, hess, hess_val):
    self.hess = hess
    self.x = x
    self.hess_op = hess_val

  def set_value(self, x):
    self.x = x
    self.hess_op = None

  def _eval(self, args):
    if self.hess_op is None:
      self.hess_op = self.hess(self.x, *args)

  def product(self, args, direction, grad=None):
    self._eval(args)
    return self.hess_op(direction)


class FindiffHessian(Hessian):
  def __init__(self, deriv):
    self.deriv = deriv
    self.hess = None

  def set_value(self, x):
    self.x = x
    self.hess = None

  def _eval(self, grad, args):
    if self.hess is None:
      self.hess = self.deriv(self.x, args, grad)

  def product(self, args, direction, grad=None):
    self._eval(grad, args)
    return np.dot(self.hess, direction)


class GenericHessian(Hessian):
  def __init__(self, hess):
    self.hess = hess
    self.impl = None

  def product(self, args, direction, grad=None):
    hess_val = self.hess(self.x, *args)

    self.impl = _create_callable_hessian(self.x,
                                         self.hess,
                                         hess_val)

    def set_value(x):
      self.impl.set_value(x)

    def product(args, direction, grad=None):
      return self.impl.product(args, direction, grad)

    self.set_value = set_value
    self.product = product

    return self.impl.product(args, direction, grad)

  def set_value(self, x):
    self.x = x


def _create_callable_hessian(x, hess, hess_val):
  if sps.issparse(hess_val):
    return SparseHessian(x, hess, hess_val)

  elif isinstance(hess_val, LinearOperator):
    return HessianOperator(x, hess, hess_val)
  else:
    return DenseHessian(x, hess, hess_val)

def create_hessian(grad, hess, hessp):

  if hess and hessp:
    raise ValueError("Specified both Hessian and Hessian product")

  if hessp:
    return HessianProduct(hessp)

  if hess in FD_METHODS:
    deriv = create_derivative(grad, hess)
    return FindiffHessian(deriv)

  elif callable(hess):
    return GenericHessian(hess)

  raise ValueError("Invalid Hessian")



import logging

cimport libc.time

cdef void sleqp_python_handler(csleqp.SLEQP_LOG_LEVEL level,
                               libc.time.time_t time,
                               const char* message):
  levels = {
    csleqp.SLEQP_LOG_DEBUG: logging.DEBUG,
    csleqp.SLEQP_LOG_ERROR: logging.ERROR,
    csleqp.SLEQP_LOG_WARN: logging.WARNING,
    csleqp.SLEQP_LOG_INFO: logging.INFO,
  }

  assert level in levels

  msg = message.decode("UTF-8", "replace")

  sleqp_logger.log(levels[level], msg)

cdef void sleqp_python_handler_nogil(csleqp.SLEQP_LOG_LEVEL level,
                                     libc.time.time_t time,
                                     const char* message) nogil:
  with gil:
    sleqp_python_handler(level, time, message)


cdef update_log_handler():
  if release_gil:
    csleqp.sleqp_log_set_handler(sleqp_python_handler_nogil)
  else:
    csleqp.sleqp_log_set_handler(sleqp_python_handler)


class SleqpLogger(logging.Logger):
  def __init__(self, name, level=logging.NOTSET):
    super(SleqpLogger, self).__init__(name, level)

    update_log_handler()


  def setLevel(self, level):
    super(SleqpLogger, self).setLevel(level)

    levels = {
      logging.DEBUG: csleqp.SLEQP_LOG_DEBUG,
      logging.CRITICAL: csleqp.SLEQP_LOG_ERROR,
      logging.ERROR: csleqp.SLEQP_LOG_ERROR,
      logging.WARNING: csleqp.SLEQP_LOG_WARN,
      logging.INFO: csleqp.SLEQP_LOG_INFO,
      logging.NOTSET: csleqp.SLEQP_LOG_DEBUG
    }

    assert level in levels

    csleqp.sleqp_log_set_level(levels[level])


_logging_class = logging.getLoggerClass()
logging.setLoggerClass(SleqpLogger)

sleqp_logger = logging.getLogger('sleqp')

logging.setLoggerClass(_logging_class)



#cython: language_level=3

cimport numpy as np

from scipy.sparse import csc_matrix

cdef class MatrixRef:
  cdef csleqp.SleqpMat* matrix

  cdef _release(self):
    if self.matrix:
      csleqp_call(csleqp.sleqp_mat_release(&self.matrix))

  cdef _set_matrix(self, csleqp.SleqpMat* matrix):
    self._release()

    csleqp_call(csleqp.sleqp_mat_capture(matrix))

    self.matrix = matrix

  def __dealloc__(self):
    self._release()

cdef object sleqp_sparse_matrix_to_scipy(csleqp.SleqpMat* _matrix):
  cdef MatrixRef matrix_ref = MatrixRef()

  matrix_ref._set_matrix(_matrix)

  num_cols = csleqp.sleqp_mat_num_cols(_matrix)

  num_rows = csleqp.sleqp_mat_num_rows(_matrix)

  nnz = csleqp.sleqp_mat_nnz(_matrix)

  data = np.asarray(<double[:nnz]> csleqp.sleqp_mat_data(_matrix))

  indices = np.asarray(<int[:nnz]> csleqp.sleqp_mat_rows(_matrix))

  indptr = np.asarray(<int[:num_cols + 1]> csleqp.sleqp_mat_cols(_matrix))

  matrix = csc_matrix((data, indices, indptr), shape=(num_rows, num_cols))

  matrix._ref = matrix_ref

  return matrix



import numpy as np

from sleqp._func import create_func


class ConstraintFunc:

  def __init__(self, num_variables, constraints):
    self.num_variables = num_variables
    self.funcs = self._create_funcs(constraints)

    self.num_constraints = sum(func.dimension for func in self.funcs)

    self.shape = (self.num_constraints,
                  self.num_variables)

  def _create_func(self, constraint):

      try:
        fun = constraint.fun
        jac = constraint.jac
        hess = constraint.hess
        dim = len(constraint.lb)
        return create_func(fun, jac, hess=hess, dimension=dim)

      except AttributeError:
        pass

      fun = constraint['fun']
      jac = constraint.get('jac')
      hess = constraint.get('hess')
      hessp = constraint.get('hessp')

      return create_func(fun, jac, hess, hessp)

  def _create_funcs(self, constraints):
    funcs = []

    for constraint in constraints:
      funcs.append(self._create_func(constraint))

    return funcs

  def set_value(self, x):
    self.x = x

    for func in self.funcs:
      func.set_value(self.x)

  def val(self, args=()):
    values = []

    for i, func in enumerate(self.funcs):
      values.append(func.val(args))

    if values:
      values = np.hstack(values)
    else:
      values = np.array([])

    assert values.shape == (self.num_constraints,)

    return values

  def jac(self, args=()):
    jac = []

    for i, func in enumerate(self.funcs):
      func_grad = func.grad(args)
      if func_grad.ndim == 1:
        func_grad = func_grad[np.newaxis, :]
      assert func_grad.shape == (func.dimension, self.num_variables)
      jac.append(func_grad)

    if jac:
      jac = np.vstack(jac)
    else:
      return np.zeros(self.shape)

    return jac

  def hess_prod(self, direction, duals, args=()):
    assert direction.shape == (self.num_variables,)
    assert duals.shape == (self.num_constraints,)

    product = np.zeros_like(direction)

    offset = 0

    for i, func in enumerate(self.funcs):
      dim = func.dimension
      func_duals = duals[offset:offset + dim]

      func_product = func.hess_prod(direction, args)

      scaled_product = np.dot(func_duals,
                              func_product[np.newaxis,:])

      if scaled_product.ndim > 1:
        scaled_product = scaled_product[0, :]

      product += scaled_product

      offset += dim

    return product


def create_constraint_func(num_variables,
                           constraints):
  return ConstraintFunc(num_variables, constraints)



#cython: language_level=3

version = (csleqp.SLEQP_VERSION_MAJOR,
           csleqp.SLEQP_VERSION_MINOR,
           csleqp.SLEQP_VERSION_PATCH)

git_branch = csleqp.SLEQP_GIT_BRANCH
git_commit_hash = csleqp.SLEQP_GIT_COMMIT_HASH



#cython: language_level=3

cdef store_func_exc(func_obj, exception):
  func_obj.call_exception = exception

cdef csleqp.SLEQP_RETCODE sleqp_func_set(csleqp.SleqpFunc* func,
                                         csleqp.SleqpVec* x,
                                         csleqp.SLEQP_VALUE_REASON reason,
                                         csleqp.bool* reject,
                                         void* func_data):

  try:
    x_array = _readonly(sleqp_sparse_vec_to_array(x))

    func_obj = (<object> func_data)

    do_reject = func_obj.set_value(x_array, ValueReason(reason))

    if do_reject is True:
      reject[0] = True
      return csleqp.SLEQP_OKAY
    else:
      reject[0] = False

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_func_set_nogil(csleqp.SleqpFunc* func,
                                               csleqp.SleqpVec* x,
                                               csleqp.SLEQP_VALUE_REASON reason,
                                               csleqp.bool* reject,
                                               void* func_data) nogil:
  with gil:
    return sleqp_func_set(func,
                          x,
                          reason,
                          reject,
                          func_data)


cdef csleqp.SLEQP_RETCODE sleqp_func_nonzeros(csleqp.SleqpFunc* func,
                                              int* obj_grad_nnz,
                                              int* cons_val_nnz,
                                              int* cons_jac_nnz,
                                              int* hess_prod_nnz,
                                              void* func_data):

  try:
    func_obj = (<object> func_data)

    def try_call(f):
      try:
        return f()
      except AttributeError:
        return csleqp.SLEQP_NONE

    obj_grad_nnz[0] = try_call(lambda : func_obj.obj_grad_nnz())
    cons_val_nnz[0] = try_call(lambda : func_obj.cons_val_nnz())
    cons_jac_nnz[0] = try_call(lambda : func_obj.cons_jac_nnz())
    hess_prod_nnz[0] = try_call(lambda : func_obj.hess_prod_nnz())

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_func_nonzeros_nogil(csleqp.SleqpFunc* func,
                                                    int* obj_grad_nnz,
                                                    int* cons_val_nnz,
                                                    int* cons_jac_nnz,
                                                    int* hess_prod_nnz,
                                                    void* func_data) nogil:
  with gil:
    return sleqp_func_nonzeros(func,
                               obj_grad_nnz,
                               cons_val_nnz,
                               cons_jac_nnz,
                               hess_prod_nnz,
                               func_data)


cdef csleqp.SLEQP_RETCODE sleqp_func_obj_val(csleqp.SleqpFunc* func,
                                             double* obj_val,
                                             void* func_data):
  try:

    func_obj = (<object> func_data)

    result = func_obj.obj_val()
    assert result is not None, "obj_val() returned 'None'"
    obj_val[0] = result

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_func_obj_val_nogil(csleqp.SleqpFunc* func,
                                                   double* obj_val,
                                                   void* func_data) nogil:
  with gil:
    return sleqp_func_obj_val(func,
                              obj_val,
                              func_data)


cdef csleqp.SLEQP_RETCODE sleqp_func_obj_grad(csleqp.SleqpFunc* func,
                                              csleqp.SleqpVec* obj_grad,
                                              void* func_data):
  cdef int num_vars
  try:

    func_obj = (<object> func_data)

    num_vars = csleqp.sleqp_func_num_vars(func)

    grad_array = func_obj.obj_grad()

    assert grad_array is not None, "obj_grad() returned 'None'"
    assert grad_array.ndim == 1, "Gradient must be a vector"
    assert grad_array.size == num_vars, "Gradient has wrong size"

    csleqp_call(array_to_sleqp_sparse_vec(grad_array, obj_grad))

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_func_obj_grad_nogil(csleqp.SleqpFunc* func,
                                                    csleqp.SleqpVec* obj_grad,
                                                    void* func_data) nogil:
  with gil:
    return sleqp_func_obj_grad(func,
                               obj_grad,
                               func_data)


cdef csleqp.SLEQP_RETCODE sleqp_func_cons_val(csleqp.SleqpFunc* func,
                                              csleqp.SleqpVec* cons_vals,
                                              void* func_data):
  cdef int num_cons
  try:

    func_obj = (<object> func_data)

    num_cons = csleqp.sleqp_func_num_cons(func)

    cons_array = func_obj.cons_vals()

    assert cons_array is not None, "cons_vals() returned 'None'"
    assert cons_array.ndim == 1, "Constraint values must be a vector"
    assert cons_array.size == num_cons, "Constraint values have wrong size"

    csleqp_call(array_to_sleqp_sparse_vec(cons_array, cons_vals))

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_func_cons_val_nogil(csleqp.SleqpFunc* func,
                                                    csleqp.SleqpVec* cons_vals,
                                                    void* func_data) nogil:
  with gil:
    return sleqp_func_cons_val(func,
                               cons_vals,
                               func_data)


cdef csleqp.SLEQP_RETCODE sleqp_func_cons_jac(csleqp.SleqpFunc* func,
                                              csleqp.SleqpMat* cons_jac,
                                              void* func_data):
  cdef int num_vars
  cdef int num_cons
  try:

    func_obj = (<object> func_data)

    num_vars = csleqp.sleqp_func_num_vars(func)
    num_cons = csleqp.sleqp_func_num_cons(func)

    cons_jac_mat = func_obj.cons_jac()

    expected_shape = (num_cons, num_vars)

    assert cons_jac_mat is not None, "cons_jac() returned 'None'"
    assert cons_jac_mat.ndim == 2, "Constraint Jacobian must be a matrix"
    assert cons_jac_mat.shape == expected_shape, "Constraint Jacobian has wrong shape"

    csleqp_call(matrix_to_sleqp_sparse_matrix(cons_jac_mat, cons_jac))


  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_func_cons_jac_nogil(csleqp.SleqpFunc* func,
                                                    csleqp.SleqpMat* cons_jac,
                                                    void* func_data) nogil:
  with gil:
    return sleqp_func_cons_jac(func,
                               cons_jac,
                               func_data)


cdef csleqp.SLEQP_RETCODE sleqp_func_hess_prod(csleqp.SleqpFunc* func,
                                               const csleqp.SleqpVec* direction,
                                               const csleqp.SleqpVec* cons_dual,
                                               csleqp.SleqpVec* product,
                                               void* func_data):

  cdef int num_vars

  try:

    assert cons_dual
    assert direction
    assert product

    func_obj = (<object> func_data)

    num_vars = csleqp.sleqp_func_num_vars(func)

    direction_array = _readonly(sleqp_sparse_vec_to_array(direction))
    cons_dual_array = _readonly(sleqp_sparse_vec_to_array(cons_dual))

    product_array = func_obj.hess_prod(direction_array, cons_dual_array)

    assert product_array is not None, "hess_prod(...) returned 'None'"
    assert product_array.ndim == 1, "Hessian product must be a vector"
    assert product_array.size == num_vars, "Hessian product has wrong size"

    csleqp_call(array_to_sleqp_sparse_vec(product_array, product))

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY


cdef csleqp.SLEQP_RETCODE sleqp_func_hess_prod_nogil(csleqp.SleqpFunc* func,
                                                     const csleqp.SleqpVec* direction,
                                                     const csleqp.SleqpVec* cons_dual,
                                                     csleqp.SleqpVec* product,
                                                     void* func_data) nogil:
  with gil:
    return sleqp_func_hess_prod(func,
                                direction,
                                cons_dual,
                                product,
                                func_data)


cdef csleqp.SLEQP_RETCODE sleqp_func_free(void* func_data):
  return csleqp.SLEQP_OKAY


cdef object funcs = weakref.WeakSet()


cdef set_func_callbacks(csleqp.SleqpFuncCallbacks* callbacks):
  if release_gil:
    callbacks[0].set_value = &sleqp_func_set_nogil
    callbacks[0].nonzeros  = &sleqp_func_nonzeros_nogil
    callbacks[0].obj_val   = &sleqp_func_obj_val_nogil
    callbacks[0].obj_grad  = &sleqp_func_obj_grad_nogil
    callbacks[0].cons_val  = &sleqp_func_cons_val_nogil
    callbacks[0].cons_jac  = &sleqp_func_cons_jac_nogil
    callbacks[0].hess_prod = &sleqp_func_hess_prod_nogil
  else:
    callbacks[0].set_value = &sleqp_func_set
    callbacks[0].nonzeros  = &sleqp_func_nonzeros
    callbacks[0].obj_val   = &sleqp_func_obj_val
    callbacks[0].obj_grad  = &sleqp_func_obj_grad
    callbacks[0].cons_val  = &sleqp_func_cons_val
    callbacks[0].cons_jac  = &sleqp_func_cons_jac
    callbacks[0].hess_prod = &sleqp_func_hess_prod

  callbacks.func_free = &sleqp_func_free


cdef update_func_callbacks():
  cdef _Func func
  cdef csleqp.SleqpFuncCallbacks callbacks

  set_func_callbacks(&callbacks)

  for obj in funcs:
    func = <_Func> obj

    if not func.cfunc:
      continue

    csleqp_call(csleqp.sleqp_func_set_callbacks(func.cfunc,
                                                &callbacks))


cdef class _Func:
  cdef csleqp.SleqpFunc* cfunc
  cdef object __weakref__

  def __cinit__(self):
    self.cfunc = NULL

  cdef void set_func(self, csleqp.SleqpFunc* cfunc):
    self.release()
    self.cfunc = cfunc
    self.capture()

  cdef void capture(self):
    if self.cfunc != NULL:
      csleqp_call(csleqp.sleqp_func_capture(self.cfunc))

  cdef void release(self):
    if self.cfunc != NULL:
      csleqp_call(csleqp.sleqp_func_release(&self.cfunc))

  def __dealloc__(self):
    self.release()


cdef csleqp.SLEQP_RETCODE create_func(csleqp.SleqpFunc** cfunc,
                                      object func,
                                      int num_vars,
                                      int num_cons):
  cdef csleqp.SleqpFuncCallbacks callbacks
  cdef _Func ofunc = _Func()
  cdef csleqp.SLEQP_RETCODE retcode

  assert func is not None

  set_func_callbacks(&callbacks)

  retcode = csleqp.sleqp_func_create(cfunc,
                                     &callbacks,
                                     num_vars,
                                     num_cons,
                                     <void*> func)

  if cfunc[0] != NULL:
    ofunc.set_func(cfunc[0])
    funcs.add(ofunc)

  return retcode



cdef class HessianStruct:
  cdef csleqp.SleqpHessStruct* hess_struct
  cdef dict __dict__
  cdef _Func _func

  def __cinit__(self, _Func func):
    self._func = func
    self.hess_struct = csleqp.sleqp_func_hess_struct(func.cfunc)
    csleqp_call(csleqp.sleqp_hess_struct_capture(self.hess_struct))

  def __dealloc__(self):
    csleqp_call(csleqp.sleqp_hess_struct_release(&self.hess_struct))

  def clear(self) -> None:
    csleqp_call(csleqp.sleqp_hess_struct_clear(self.hess_struct))

  def push(self, int end) -> None:
    csleqp_call(csleqp.sleqp_hess_struct_push_block(self.hess_struct, end))

  @property
  def num_blocks(self) -> int:
    return csleqp.sleqp_hess_struct_num_blocks(self.hess_struct)

  def block_range(self, int block) -> collections.abc.Iterable[int]:
    cdef int begin = 0
    cdef int end = 0

    csleqp_call(csleqp.sleqp_hess_struct_block_range(self.hess_struct,
                                                     block,
                                                     &begin,
                                                     &end))

    return range(begin, end)

  @property
  def block_ranges(self) -> typing.Iterable[collections.abc.Iterable[int]]:

    num_blocks = self.num_blocks

    for block in range(num_blocks):
      yield self.block_range(block)

  @property
  def linear_range(self) -> collections.abc.Iterable[int]:
    cdef int begin = 0
    cdef int end = 0

    csleqp_call(csleqp.sleqp_hess_struct_lin_range(self.hess_struct,
                                                   &begin,
                                                   &end))

    return range(begin, end)

  def __str__(self):
    blocks = ', '.join(('[{0}, {1})'.format(block_range.start,
                                            block_range.stop)
                        for block_range in self.block_ranges))

    return 'HessianStruct: [{0}]'.format(blocks)



from dataclasses import dataclass

import numpy as np
import scipy.sparse


@dataclass
class LinearConstraints:
  lb: object = None
  coeffs: object = None
  ub: object = None


def convert_to_sparse(array):
  (m, n) = array.shape

  mat = scipy.sparse.dok_matrix((m, n))

  for i in range(m):
    for j in range(n):
      value = array[i, j]

      if value:
        mat[i, j] = value

  return mat


def create_linear_constraints(constraints):

  if not(constraints):
    return LinearConstraints()

  linear_lb = np.hstack([constraint.lb for constraint in constraints])

  linear_ub = np.hstack([constraint.ub for constraint in constraints])

  matrices = []

  for constraint in constraints:
    matrix = constraint.A
    if not(scipy.sparse.issparse(matrix)):
      matrix = convert_to_sparse(matrix)

    matrices.append([matrix])

  matrix = scipy.sparse.bmat(matrices)

  return LinearConstraints(lb=linear_lb, coeffs=matrix, ub=linear_ub)



#cython: language_level=3

class Array(np.ndarray):
  """
  Thin wrapper around a :class:`numpy.ndarray`, storing a
  reference to prevent the destruction of
  an underlying object
  """
  def __new__(cls, input_array, token=None):
    obj = np.asarray(input_array).view(cls)
    obj._token = token
    return obj

  def __array_finalize__(self, obj):
    if obj is None:
      return
    self._token = getattr(obj, '_token', None)

cdef class Scaling:
  """
  Scaling of a problem, consisting of weights (powers of 2) for
  variables, constraints, and objective. Scaling weights can also be
  given as nominal values, i.e., values which are transformed to ~1 by
  the scaling.

  """
  cdef dict __dict__
  cdef csleqp.SleqpScaling* scaling
  cdef csleqp.SleqpVec* gradient
  cdef csleqp.SleqpMat* cons_jac

  def __cinit__(self,
                int num_vars,
                int num_cons,
                *args,
                **keywords):

    csleqp_call(csleqp.sleqp_scaling_create(&self.scaling,
                                            num_vars,
                                            num_cons))

    csleqp_call(csleqp.sleqp_vec_create_empty(&self.gradient,
                                                        num_vars))

    csleqp_call(csleqp.sleqp_mat_create(&self.cons_jac,
                                                  num_cons,
                                                  num_vars,
                                                  0))

  @property
  def num_vars(self):
    """
    Number of variables in the scaling
    """
    return csleqp.sleqp_scaling_num_vars(self.scaling)

  @property
  def num_cons(self):
    """
    Number of constraints in the scaling
    """
    return csleqp.sleqp_scaling_num_cons(self.scaling)

  def __dealloc__(self):
    csleqp_call(csleqp.sleqp_mat_release(&self.cons_jac))
    csleqp_call(csleqp.sleqp_vec_free(&self.gradient))
    csleqp_call(csleqp.sleqp_scaling_release(&self.scaling))

  def __str__(self):
    val = "Scaling(num_variables={0}, num_constraints={1})\n".format(self.num_vars,
                                                                     self.num_cons)

    val += "Objective weight: {0}\n".format(self.obj_weight)

    val += "Variable weights: {0}\n".format(self.variable_weights)

    val += "Constraint weights: {0}\n".format(self.variable_weights)

    return val

  @staticmethod
  def weights_to_nominal_values(weights):
    """
    Convert weights to nominal values
    """
    return np.ldexp(np.ones(weights.shape), weights)

  @staticmethod
  def nominal_values_to_weights(nominal_values):
    """
    Convert nominal values to weights
    """
    return np.frexp(nominal_values)[1]

  @staticmethod
  def nominal_value_to_weight(nominal_value):
    """
    Convert single nominal value to weight
    """
    return np.frexp(np.array([nominal_value]))[1].item()

  @staticmethod
  def weight_to_nominal_value(weight):
    """
    Convert single weight to nominal value
    """
    return np.ldexp(np.ones(1), np.array([weight])).item()

  @property
  def obj_weight(self):
    """
    Weight of objective
    """
    return csleqp.sleqp_scaling_obj_weight(self.scaling)

  @obj_weight.setter
  def obj_weight(self, value):
    csleqp_call(csleqp.sleqp_scaling_set_obj_weight(self.scaling,
                                                    value))

  @property
  def var_weights(self):
    """
    Weights of variables
    """
    length = self.num_vars
    cdef int[:] values = <int[:length]> csleqp.sleqp_scaling_var_weights(self.scaling)

    array = np.asarray(values)
    array.flags.writeable = False

    return Array(array, self)

  @var_weights.setter
  def var_weights(self, values):
    assert values.shape == (self.num_vars,)
    assert values.dtype == np.int64

    for i in range(self.num_vars):
      self.set_variable_weight(i, values[i])

  @property
  def cons_weights(self):
    """
    Weights of constraints
    """
    length = self.num_cons
    cdef int[:] values = <int[:length]> csleqp.sleqp_scaling_cons_weights(self.scaling)

    array = np.asarray(values)
    array.flags.writeable = False

    return Array(array, self)

  @cons_weights.setter
  def cons_weights(self, values):
    assert values.shape == (self.num_cons,)
    assert values.dtype == np.int64

    for i in range(self.num_cons):
      self.set_constraint_weight(i, values[i])

  def set_obj_weight_from_nominal(self, nominal_value):
    csleqp_call(csleqp.sleqp_scaling_set_obj_weight_from_nominal(self.scaling,
                                                                 nominal_value))

  def set_var_weight(self, int index, int weight):
    csleqp_call(csleqp.sleqp_scaling_set_var_weight(self.scaling,
                                                    index,
                                                    weight))

  def set_var_weights_from_nominal(self, nominal_array):
    assert nominal_array.shape == (self.num_vars,)
    cdef double[:] nominal_values = nominal_array
    csleqp_call(csleqp.sleqp_scaling_set_var_weights_from_nominal(self.scaling,
                                                                  &nominal_values[0]))

  def set_var_weight_from_nominal(self, int index, float nominal_value):
    csleqp_call(csleqp.sleqp_scaling_set_var_weight_from_nominal(self.scaling,
                                                                 index,
                                                                 nominal_value))

  def set_cons_weight(self, int index, int weight):
    csleqp_call(csleqp.sleqp_scaling_set_cons_weight(self.scaling,
                                                     index,
                                                     weight))

  def set_cons_weights_from_nominal(self, nominal_array):
    assert nominal_array.shape == (self.num_cons,)
    cdef double[:] nominal_values = nominal_array
    csleqp_call(csleqp.sleqp_scaling_set_cons_weights_from_nominal(self.scaling,
                                                                   &nominal_values[0]))

  def set_cons_weight_from_nominal(self, int index, float nominal_value):
    csleqp_call(csleqp.sleqp_scaling_set_cons_weight_from_nominal(self.scaling,
                                                                  index,
                                                                  nominal_value))

  def set_from_obj_grad(self,
                        np.ndarray grad_array,
                        float eps):
    csleqp_call(array_to_sleqp_sparse_vec(grad_array,
                                          self.gradient))

    csleqp_call(csleqp.sleqp_obj_scaling_from_grad(self.scaling,
                                                   self.gradient,
                                                   eps))

  def set_from_cons_jac(self,
                        object mat,
                        float eps):
    csleqp_call(matrix_to_sleqp_sparse_matrix(mat,
                                              self.cons_jac))

    csleqp_call(csleqp.sleqp_scaling_from_cons_jac(self.scaling,
                                                   self.cons_jac,
                                                   eps))



#cython: language_level=3

from libc.stdlib cimport malloc, free

cdef object solvers = weakref.WeakSet()

cdef class Solver:
  """
  Solver class for NLP problems
  """
  cdef dict __dict__

  cdef object __weakref__

  cdef csleqp.SleqpSolver* solver
  cdef object problem
  cdef csleqp.SleqpVec* residuals

  cdef list callback_handles

  def __cinit__(self,
                object problem,
                np.ndarray primal,
                Scaling scaling=None):

    cdef csleqp.SleqpVec* primal_vec
    cdef _Problem _problem = <_Problem> problem._get_problem()

    self.callback_handles = []

    csleqp_call(csleqp.sleqp_vec_create_empty(&primal_vec,
                                                        problem.num_vars))

    csleqp_call(csleqp.sleqp_vec_create_empty(&self.residuals, 0))

    array_to_sleqp_sparse_vec(primal, primal_vec)

    csleqp_call(csleqp.sleqp_solver_create(&self.solver,
                                           _problem.cproblem,
                                           primal_vec,
                                           scaling.scaling if scaling is not None else NULL))

    csleqp_call(csleqp.sleqp_vec_free(&primal_vec))

    self.problem = problem

    solvers.add(self)

  def __dealloc__(self):
    assert self.solver

    csleqp_call(csleqp.sleqp_solver_release(&self.solver))

    csleqp_call(csleqp.sleqp_vec_free(&self.residuals))

  cdef _solve(self,
              int max_num_iterations,
              double time_limit):

    cdef csleqp.SleqpSolver* solver = self.solver
    cdef csleqp.SLEQP_RETCODE retcode = csleqp.SLEQP_OKAY
    cdef csleqp.SLEQP_ERROR_TYPE error_type

    self.problem.func.call_exception = None

    for obj in self.callback_handles:
      (<CallbackHandle> obj).call_exception = None

    if release_gil:
      with nogil:
        retcode = csleqp.sleqp_solver_solve(self.solver,
                                            max_num_iterations,
                                            time_limit)
    else:
        retcode = csleqp.sleqp_solver_solve(self.solver,
                                            max_num_iterations,
                                            time_limit)

    if retcode == csleqp.SLEQP_OKAY:
      return

    exception = Exception("Error during solution process")
    error_type = csleqp.sleqp_error_type()

    if error_type == csleqp.SLEQP_FUNC_EVAL_ERROR:
      call_exception = self.problem.func.call_exception
      inner_exception = _get_exception()

      if call_exception:
        inner_exception.__cause__ = call_exception
        self.problem.func.call_exception = None

      raise exception from inner_exception

    elif error_type == csleqp.SLEQP_CALLBACK_ERROR:
      callback_exception = CallbackError()
      call_exception = None

      for obj in self.callback_handles:
        call_exception = (<CallbackHandle> obj).call_exception
        if call_exception is not None:
          (<CallbackHandle> obj).call_exception = None
          break

      if call_exception:
        callback_exception.__cause__ = call_exception
      raise exception from callback_exception

    raise exception from _get_exception()

  def info(self) -> str:
    return str(csleqp.sleqp_solver_info(self.solver))

  def solve(self,
            max_num_iterations: int = None,
            time_limit: float = None) -> None:
    """
    Solve the underlying NLP.

    :param max_num_iterations: Iteration limit, `None` for unlimited
    :type value: int

    :param time_limit: Time limit, `None` for unlimited
    :type value: float
    """

    cdef int max_it = csleqp.SLEQP_NONE
    cdef double time = csleqp.SLEQP_NONE

    if max_num_iterations is not None:
      max_it = max_num_iterations

    if time_limit is not None:
      time = time_limit

    self._solve(max_it, time)

  @property
  def status(self) -> Status:
    """
    Returns status code of last solve
    """
    return Status(csleqp.sleqp_solver_status(self.solver))

  def abort(self):
    """
    Aborts the solver as soon as the current iteration
    finishes.
    """
    csleqp_call(csleqp.sleqp_solver_abort(self.solver))

  @property
  def iterations(self) -> int:
    return csleqp.sleqp_solver_iterations(self.solver)

  @property
  def elapsed_seconds(self) -> float:
    return csleqp.sleqp_solver_elapsed_seconds(self.solver)

  @property
  def violated_cons(self) -> set:
    num_constraints =  self.problem.num_cons

    cdef int *violated_cons = <int *> malloc(num_constraints * sizeof(double))
    cdef int num_violated_cons
    cdef csleqp.SleqpIterate* iterate

    try:
      csleqp_call(csleqp.sleqp_solver_solution(self.solver, &iterate))

      csleqp_call(csleqp.sleqp_solver_violated_constraints(self.solver,
                                                           iterate,
                                                           violated_cons,
                                                           &num_violated_cons))

      violated = set()

      for i in range(num_violated_cons):
        violated.add(violated_cons[i])

      return violated

    finally:
      free(violated_cons)

  @property
  def solution(self) -> Iterate:
    """
    Solution iterate of last call to solve()
    """
    cdef csleqp.SleqpIterate* _iterate
    cdef Iterate iterate = Iterate(_create=True)

    csleqp_call(csleqp.sleqp_solver_solution(self.solver, &_iterate))

    iterate._set_iterate(_iterate)

    return iterate


  def add_callback(self, event, function):
    """
    Adds a callback function to the solver
    """
    cdef CallbackHandle callback_handle = CallbackHandle(self, event, function)

    csleqp_call(get_callback_function_pointer(event,
                                              &callback_handle.function_pointer))

    csleqp_call(csleqp.sleqp_solver_add_callback(self.solver,
                                                 callback_handle.event.value,
                                                 callback_handle.function_pointer,
                                                 <void*> callback_handle))

    self.callback_handles.append(callback_handle)

    return callback_handle

  def remove_callback(self, CallbackHandle callback_handle):

    csleqp_call(csleqp.sleqp_solver_remove_callback(self.solver,
                                                    callback_handle.event.value,
                                                    callback_handle.function_pointer,
                                                    <void*> callback_handle))

    self.callback_handles.remove(callback_handle)

  cdef update_callbacks(self):
    cdef CallbackHandle callback_handle

    for obj in self.callback_handles:
      callback_handle = <CallbackHandle> obj

      csleqp_call(csleqp.sleqp_solver_remove_callback(self.solver,
                                                      callback_handle.event.value,
                                                      callback_handle.function_pointer,
                                                      <void*> callback_handle))

      csleqp_call(get_callback_function_pointer(callback_handle.event,
                                                &callback_handle.function_pointer))

      csleqp_call(csleqp.sleqp_solver_add_callback(self.solver,
                                                   callback_handle.event.value,
                                                   callback_handle.function_pointer,
                                                   <void*> callback_handle))

  @property
  def states(self):
    """
    Return state of the solver as dict
    """
    stat_residuals = None

    csleqp_call(csleqp.sleqp_solver_vec_state(self.solver,
                                              csleqp.SLEQP_SOLVER_STATE_VEC_SCALED_STAT_RESIDUALS,
                                              self.residuals))
    stat_residuals = sleqp_sparse_vec_to_array(self.residuals)

    csleqp_call(csleqp.sleqp_solver_vec_state(self.solver,
                                              csleqp.SLEQP_SOLVER_STATE_VEC_SCALED_FEAS_RESIDUALS,
                                              self.residuals))
    feas_residuals = sleqp_sparse_vec_to_array(self.residuals)

    csleqp_call(csleqp.sleqp_solver_vec_state(self.solver,
                                              csleqp.SLEQP_SOLVER_STATE_VEC_SCALED_CONS_SLACK_RESIDUALS,
                                              self.residuals))
    cons_slack_residuals = sleqp_sparse_vec_to_array(self.residuals)

    csleqp_call(csleqp.sleqp_solver_vec_state(self.solver,
                                              csleqp.SLEQP_SOLVER_STATE_VEC_SCALED_VAR_SLACK_RESIDUALS,
                                              self.residuals))
    vars_slack_residuals = sleqp_sparse_vec_to_array(self.residuals)

    return {
      SolverState.TrustRadius:              self._get_solver_real_state(csleqp.SLEQP_SOLVER_STATE_REAL_TRUST_RADIUS),
      SolverState.LPTrustRadius:            self._get_solver_real_state(csleqp.SLEQP_SOLVER_STATE_REAL_LP_TRUST_RADIUS),
      SolverState.ScaledFuncVal:            self._get_solver_real_state(csleqp.SLEQP_SOLVER_STATE_REAL_SCALED_OBJ_VAL),
      SolverState.ScaledMeritVal:           self._get_solver_real_state(csleqp.SLEQP_SOLVER_STATE_REAL_SCALED_MERIT_VAL),
      SolverState.ScaledFeasRes:            self._get_solver_real_state(csleqp.SLEQP_SOLVER_STATE_REAL_SCALED_FEAS_RES),
      SolverState.ScaledStatRes:            self._get_solver_real_state(csleqp.SLEQP_SOLVER_STATE_REAL_SCALED_STAT_RES),
      SolverState.ScaledSlackRes:           self._get_solver_real_state(csleqp.SLEQP_SOLVER_STATE_REAL_SCALED_SLACK_RES),
      SolverState.PenaltyParameter:         self._get_solver_real_state(csleqp.SLEQP_SOLVER_STATE_REAL_PENALTY_PARAM),
      SolverState.MinRayleigh:              self._get_solver_real_state(csleqp.SLEQP_SOLVER_STATE_REAL_MIN_RAYLEIGH),
      SolverState.MaxRayleigh:              self._get_solver_real_state(csleqp.SLEQP_SOLVER_STATE_REAL_MAX_RAYLEIGH),
      SolverState.LastStepType:             StepType(self._get_solver_int_state(csleqp.SLEQP_SOLVER_STATE_INT_LAST_STEP_TYPE)),
      SolverState.Iteration:                self._get_solver_int_state(csleqp.SLEQP_SOLVER_STATE_INT_ITERATION),
      SolverState.LastStepOnBoundary:       bool(self._get_solver_int_state(csleqp.SLEQP_SOLVER_STATE_INT_LAST_STEP_ON_BDRY)),
      SolverState.ScaledStatResiduals:      stat_residuals,
      SolverState.ScaledFeasResiduals:      feas_residuals,
      SolverState.ScaledConsSlackResiduals: cons_slack_residuals,
      SolverState.ScaledVarSlackResiduals:  vars_slack_residuals
    }

  cpdef double _get_solver_real_state(self, int state):
    cdef double value = 0.
    csleqp_call(csleqp.sleqp_solver_real_state(self.solver,
                                               <csleqp.SLEQP_SOLVER_STATE_REAL> state,
                                               &value))
    return value

  cpdef int _get_solver_int_state(self, int state):
    cdef int value = 0
    csleqp_call(csleqp.sleqp_solver_int_state(self.solver,
                                              <csleqp.SLEQP_SOLVER_STATE_INT> state,
                                              &value))
    return value


cdef update_solver_callbacks():
  cdef Solver solver
  for obj in solvers:
    solver = <Solver> obj

    solver.update_callbacks()



import numpy as np

from sleqp._derivative import create_derivative
from sleqp._hessian import create_hessian

class PointFunc:
  def __init__(self, fun, deriv, hessian, dimension=1):
    self.x = None
    self.fun = fun
    self.deriv = deriv
    self.hessian = hessian
    self.dimension = dimension

  def set_value(self, x):
    self.x = x
    self.fval = None
    self.gval = None

    if self.hessian:
      self.hessian.set_value(x)

  def val(self, args=()):
    if self.fval is None:
      self.fval = self.fun(self.x, *args)
    return self.fval

  def grad(self, args=()):
    if self.gval is None:
      self.gval = self.deriv(self.x, args, self.fval)
      self.gval = np.atleast_1d(self.gval)
    return self.gval

  def hess_prod(self, direction, args=()):
    return self.hessian.product(args, direction)


class CombinedFunc:
  def __init__(self, fun, hessian, dimension=1):
    self.x = None
    self.fun = fun
    self.hessian = hessian
    self.dimension = dimension

  def set_value(self, x):
    self.x = x
    self.fval = None
    self.gval = None

    if self.hessian:
      self.hessian.set_value(x)

  def _eval(self, x, args=()):
    (self.fval, self.gval) = self.fun(x, *args)
    self.gval = np.atleast_1d(self.gval)

  def val(self, args=()):
    if self.fval is None:
      self._eval(self.x, *args)

    return self.fval

  def grad(self, args=()):
    if self.gval is None:
      self._eval(self,x, *args)

    return self.gval

  def hess_prod(self, direction, args=()):
    prod = self.hessian.product(args, direction)
    return np.atleast_1d(prod)


def actual_gradient(fun, grad):
  if grad is True:
    def _grad(x, *args):
      return fun(x, *args)[1]

    return _grad

  return grad

def create_func(fun, grad, hess=None, hessp=None, dimension=1):

  actual_grad = actual_gradient(fun, grad)
  hessian = None

  if hess is not None or hessp is not None:
    hessian = create_hessian(actual_grad, hess, hessp)

  if grad is True:
    return CombinedFunc(fun, hessian, dimension)

  deriv = create_derivative(fun, grad)

  return PointFunc(fun, deriv, hessian, dimension)



from inspect import currentframe, getframeinfo

cdef store_callback_exc(CallbackHandle handle, exception):
  handle.call_exception = exception

cdef class CallbackHandle:
  cdef Solver solver
  cdef object function
  cdef void* function_pointer
  cdef object event

  cdef object call_exception

  def __cinit__(self, Solver solver, object event, function):
    self.solver = solver
    self.function = function
    self.call_exception = None
    self.event = event

  def unregister(self):
    self.solver.remove_callback(self)


cdef csleqp.SLEQP_RETCODE accepted_iterate(csleqp.SleqpSolver* sol,
                                           csleqp.SleqpIterate* it,
                                           void* callback_data):
  cdef Iterate iterate = Iterate(_create=True)
  cdef CallbackHandle callback_object = None
  iterate._set_iterate(it)

  try:
    callback_object = (<CallbackHandle> callback_data)

    function = callback_object.function
    solver = callback_object.solver

    assert solver.solver == sol

    function(solver, iterate)

  except BaseException as exception:
    store_callback_exc(callback_object, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY

cdef csleqp.SLEQP_RETCODE accepted_iterate_nogil(csleqp.SleqpSolver* solver,
                                                 csleqp.SleqpIterate* iterate,
                                                 void* callback_data) nogil:
  with gil:
    return accepted_iterate(solver, iterate, callback_data)


cdef csleqp.SLEQP_RETCODE performed_iteration(csleqp.SleqpSolver* sol,
                                              void* callback_data):
  try:
    callback_object = (<CallbackHandle> callback_data)

    function = callback_object.function
    solver = callback_object.solver

    assert solver.solver == sol

    function(solver)

  except BaseException as exception:
    store_callback_exc(callback_object, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY

cdef csleqp.SLEQP_RETCODE performed_iteration_nogil(csleqp.SleqpSolver* solver,
                                                    void* callback_data) nogil:
  with gil:
    return performed_iteration(solver, callback_data)


cdef csleqp.SLEQP_RETCODE finished(csleqp.SleqpSolver* sol,
                                   csleqp.SleqpIterate* it,
                                   void* callback_data):
  cdef Iterate iterate = Iterate(_create=True)
  cdef CallbackHandle callback_object = None
  iterate._set_iterate(it)

  try:
    callback_object = (<CallbackHandle> callback_data)

    function = callback_object.function
    solver = callback_object.solver

    assert solver.solver == sol

    function(solver, iterate)

  except BaseException as exception:
    store_callback_exc(callback_object, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY

cdef csleqp.SLEQP_RETCODE finished_nogil(csleqp.SleqpSolver* solver,
                                         csleqp.SleqpIterate* iterate,
                                         void* callback_data) nogil:
  with gil:
    return finished(solver, iterate, callback_data)


cdef  get_callback_function_pointer(solver_event, void** pointer):
  if solver_event == SolverEvent.AcceptedIterate:
    if get_release_gil():
      pointer[0] = <void*> accepted_iterate_nogil
    else:
      pointer[0] = <void*> accepted_iterate
  elif solver_event == SolverEvent.PerformedIteration:
    if get_release_gil():
      pointer[0] = <void*> performed_iteration_nogil
    else:
      pointer[0] = <void*> performed_iteration
  elif solver_event == SolverEvent.Finished:
    if get_release_gil():
      pointer[0] = <void*> finished_nogil
    else:
      pointer[0] = <void*> finished
  else:
    pointer[0] = NULL
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY



#cython: language_level=3

cimport libc.stdio
cimport libc.time

cdef extern from "sleqp.h":

  ctypedef bint bool

  cdef int SLEQP_VERSION_MAJOR
  cdef int SLEQP_VERSION_MINOR
  cdef int SLEQP_VERSION_PATCH

  cdef char* SLEQP_GIT_BRANCH
  cdef char* SLEQP_GIT_COMMIT_HASH

  cdef enum:
    SLEQP_NONE

  ctypedef enum SLEQP_RETCODE:
    SLEQP_ERROR,
    SLEQP_OKAY

  ctypedef enum SLEQP_ERROR_TYPE:
    SLEQP_FAILED_ASSERTION,
    SLEQP_NOMEM,
    SLEQP_INTERNAL_ERROR,
    SLEQP_FUNC_EVAL_ERROR,
    SLEQP_CALLBACK_ERROR,
    SLEQP_MATH_ERROR,
    SLEQP_INVALID_DERIV,
    SLEQP_ILLEGAL_ARGUMENT

  ctypedef enum SLEQP_ACTIVE_STATE:
    SLEQP_INACTIVE,
    SLEQP_ACTIVE_LOWER,
    SLEQP_ACTIVE_UPPER,
    SLEQP_ACTIVE_BOTH

  ctypedef enum SLEQP_STATUS:
    SLEQP_STATUS_UNKNOWN,
    SLEQP_STATUS_RUNNING,
    SLEQP_STATUS_OPTIMAL,
    SLEQP_STATUS_INFEASIBLE,
    SLEQP_STATUS_UNBOUNDED,
    SLEQP_STATUS_ABORT_DEADPOINT,
    SLEQP_STATUS_ABORT_ITER,
    SLEQP_STATUS_ABORT_MANUAL,
    SLEQP_STATUS_ABORT_TIME

  ctypedef enum SLEQP_DERIV_CHECK:
    SLEQP_DERIV_CHECK_SKIP,
    SLEQP_DERIV_CHECK_FIRST_OBJ,
    SLEQP_DERIV_CHECK_FIRST_CONS,
    SLEQP_DERIV_CHECK_FIRST,
    SLEQP_DERIV_CHECK_SECOND_OBJ,
    SLEQP_DERIV_CHECK_SECOND_CONS,
    SLEQP_DERIV_CHECK_SECOND_EXHAUSTIVE,
    SLEQP_DERIV_CHECK_SECOND_SIMPLE

  ctypedef enum SLEQP_HESS_EVAL:
    SLEQP_HESS_EVAL_EXACT,
    SLEQP_HESS_EVAL_SR1,
    SLEQP_HESS_EVAL_SIMPLE_BFGS,
    SLEQP_HESS_EVAL_DAMPED_BFGS

  ctypedef enum SLEQP_BFGS_SIZING:
    SLEQP_BFGS_SIZING_NONE,
    SLEQP_BFGS_SIZING_CENTERED_OL

  ctypedef enum SLEQP_TR_SOLVER:
    SLEQP_TR_SOLVER_TRLIB
    SLEQP_TR_SOLVER_CG
    SLEQP_TR_SOLVER_LSQR
    SLEQP_TR_SOLVER_AUTO

  ctypedef enum SLEQP_POLISHING_TYPE:
    SLEQP_POLISHING_NONE
    SLEQP_POLISHING_ZERO_DUAL
    SLEQP_POLISHING_INACTIVE

  ctypedef enum SLEQP_STEP_RULE:
    SLEQP_STEP_RULE_DIRECT
    SLEQP_STEP_RULE_WINDOW
    SLEQP_STEP_RULE_MINSTEP

  ctypedef enum SLEQP_PARAMETRIC_CAUCHY:
    SLEQP_PARAMETRIC_CAUCHY_DISABLED
    SLEQP_PARAMETRIC_CAUCHY_COARSE
    SLEQP_PARAMETRIC_CAUCHY_FINE

  ctypedef enum SLEQP_INITIAL_TR_CHOICE:
    SLEQP_INITIAL_TR_CHOICE_NARROW
    SLEQP_INITIAL_TR_CHOICE_WIDE

  ctypedef enum SLEQP_AUG_JAC_METHOD:
    SLEQP_AUG_JAC_AUTO,
    SLEQP_AUG_JAC_STANDARD,
    SLEQP_AUG_JAC_REDUCED,
    SLEQP_AUG_JAC_DIRECT

  ctypedef enum SLEQP_LINESEARCH:
    SLEQP_LINESEARCH_EXACT
    SLEQP_LINESEARCH_APPROX

  ctypedef enum SLEQP_VALUE_REASON:
    SLEQP_VALUE_REASON_NONE,
    SLEQP_VALUE_REASON_INIT,
    SLEQP_VALUE_REASON_CHECKING_DERIV,
    SLEQP_VALUE_REASON_ACCEPTED_ITERATE,
    SLEQP_VALUE_REASON_TRYING_ITERATE,
    SLEQP_VALUE_REASON_TRYING_SOC_ITERATE,
    SLEQP_VALUE_REASON_REJECTED_ITERATE

  ctypedef enum SLEQP_DUAL_ESTIMATION_TYPE:
    SLEQP_DUAL_ESTIMATION_TYPE_LP,
    SLEQP_DUAL_ESTIMATION_TYPE_LSQ,

  ctypedef enum SLEQP_SETTINGS_REAL:
    SLEQP_SETTINGS_REAL_ZERO_EPS,
    SLEQP_SETTINGS_REAL_EPS,
    SLEQP_SETTINGS_REAL_DERIV_PERTURBATION,
    SLEQP_SETTINGS_REAL_DERIV_TOL,
    SLEQP_SETTINGS_REAL_CAUCHY_TAU,
    SLEQP_SETTINGS_REAL_CAUCHY_ETA,
    SLEQP_SETTINGS_REAL_LINESEARCH_TAU,
    SLEQP_SETTINGS_REAL_LINESEARCH_ETA,
    SLEQP_SETTINGS_REAL_LINESEARCH_CUTOFF,
    SLEQP_SETTINGS_REAL_FEAS_TOL,
    SLEQP_SETTINGS_REAL_SLACK_TOL,
    SLEQP_SETTINGS_REAL_STAT_TOL,
    SLEQP_SETTINGS_REAL_ACCEPTED_REDUCTION,
    SLEQP_SETTINGS_REAL_DEADPOINT_BOUND

  ctypedef enum SLEQP_SETTINGS_INT:
    SLEQP_SETTINGS_INT_NUM_QUASI_NEWTON_ITERATES,
    SLEQP_SETTINGS_INT_MAX_NEWTON_ITERATIONS,
    SLEQP_SETTINGS_INT_NUM_THREADS

  ctypedef enum SLEQP_SETTINGS_ENUM:
    SLEQP_SETTINGS_ENUM_DERIV_CHECK,
    SLEQP_SETTINGS_ENUM_HESS_EVAL,
    SLEQP_SETTINGS_ENUM_DUAL_ESTIMATION_TYPE,
    SLEQP_SETTINGS_ENUM_FLOAT_WARNING_FLAGS,
    SLEQP_SETTINGS_ENUM_FLOAT_ERROR_FLAGS,
    SLEQP_SETTINGS_ENUM_BFGS_SIZING,
    SLEQP_SETTINGS_ENUM_TR_SOLVER,
    SLEQP_SETTINGS_ENUM_POLISHING_TYPE,
    SLEQP_SETTINGS_ENUM_STEP_RULE,
    SLEQP_SETTINGS_ENUM_LINESEARCH,
    SLEQP_SETTINGS_ENUM_PARAMETRIC_CAUCHY,
    SLEQP_SETTINGS_ENUM_INITIAL_TR_CHOICE,
    SLEQP_SETTINGS_ENUM_AUG_JAC_METHOD

  ctypedef enum SLEQP_SETTINGS_BOOL:
    SLEQP_SETTINGS_BOOL_PERFORM_NEWTON_STEP,
    SLEQP_SETTINGS_BOOL_GLOBAL_PENALTY_RESETS,
    SLEQP_SETTINGS_BOOL_PERFORM_SOC,
    SLEQP_SETTINGS_BOOL_USE_QUADRATIC_MODEL,
    SLEQP_SETTINGS_BOOL_ENABLE_RESTORATION_PHASE,
    SLEQP_SETTINGS_BOOL_ENABLE_PREPROCESSOR,
    SLEQP_SETTINGS_BOOL_LP_RESOLVES

  ctypedef enum SLEQP_SOLVER_STATE_REAL:
    SLEQP_SOLVER_STATE_REAL_TRUST_RADIUS,
    SLEQP_SOLVER_STATE_REAL_LP_TRUST_RADIUS,
    SLEQP_SOLVER_STATE_REAL_SCALED_OBJ_VAL,
    SLEQP_SOLVER_STATE_REAL_SCALED_MERIT_VAL,
    SLEQP_SOLVER_STATE_REAL_SCALED_FEAS_RES,
    SLEQP_SOLVER_STATE_REAL_SCALED_STAT_RES,
    SLEQP_SOLVER_STATE_REAL_SCALED_SLACK_RES,
    SLEQP_SOLVER_STATE_REAL_PENALTY_PARAM,
    SLEQP_SOLVER_STATE_REAL_MIN_RAYLEIGH,
    SLEQP_SOLVER_STATE_REAL_MAX_RAYLEIGH

  ctypedef enum SLEQP_SOLVER_STATE_INT:
    SLEQP_SOLVER_STATE_INT_LAST_STEP_ON_BDRY,
    SLEQP_SOLVER_STATE_INT_ITERATION,
    SLEQP_SOLVER_STATE_INT_LAST_STEP_TYPE

  ctypedef enum SLEQP_SOLVER_STATE_VEC:
    SLEQP_SOLVER_STATE_VEC_SCALED_STAT_RESIDUALS,
    SLEQP_SOLVER_STATE_VEC_SCALED_FEAS_RESIDUALS,
    SLEQP_SOLVER_STATE_VEC_SCALED_CONS_SLACK_RESIDUALS,
    SLEQP_SOLVER_STATE_VEC_SCALED_VAR_SLACK_RESIDUALS

  ctypedef enum SLEQP_STEPTYPE:
    SLEQP_STEPTYPE_NONE,
    SLEQP_STEPTYPE_ACCEPTED,
    SLEQP_STEPTYPE_ACCEPTED_FULL,
    SLEQP_STEPTYPE_ACCEPTED_SOC,
    SLEQP_STEPTYPE_REJECTED

  ctypedef enum SLEQP_SOLVER_EVENT:
    SLEQP_SOLVER_EVENT_ACCEPTED_ITERATE,
    SLEQP_SOLVER_EVENT_PERFORMED_ITERATION,
    SLEQP_SOLVER_EVENT_FINISHED,

  ctypedef struct SleqpVec:
    double* data
    int* indices

    int dim
    int nnz
    int nnz_max

  ctypedef struct SleqpMat:
    pass

  ctypedef struct SleqpProblem:
    pass

  ctypedef struct SleqpSolver:
    pass

  ctypedef struct SleqpScaling:
    pass

  ctypedef struct SleqpSettings:
    pass

  ctypedef struct SleqpWorkingSet:
    pass

  ctypedef struct SleqpIterate:
    pass

  cdef struct SleqpFunc:
    pass

  cdef struct SleqpHessStruct:
    pass


  ctypedef enum SLEQP_LOG_LEVEL:
    SLEQP_LOG_ERROR,
    SLEQP_LOG_WARN,
    SLEQP_LOG_INFO,
    SLEQP_LOG_DEBUG

  SLEQP_ERROR_TYPE sleqp_error_type()

  const char* sleqp_error_msg()

  # Sparse vectors
  SLEQP_RETCODE sleqp_vec_create(SleqpVec** vec,
                                           int dim,
                                           int nnz_max)

  SLEQP_RETCODE sleqp_vec_create_empty(SleqpVec** vec,
                                                 int dim)

  SLEQP_RETCODE sleqp_vec_create_full(SleqpVec** vec,
                                                int dim)

  SLEQP_RETCODE sleqp_vec_fprintf(const SleqpVec* vec,
                                            libc.stdio.FILE* output)

  SLEQP_RETCODE sleqp_vec_free(SleqpVec** vec)

  SLEQP_RETCODE sleqp_vec_clear(SleqpVec* vec)

  SLEQP_RETCODE sleqp_vec_reserve(SleqpVec* vec, int nnz)

  SLEQP_RETCODE sleqp_vec_resize(SleqpVec* vec,
                                           int dim)

  SLEQP_RETCODE sleqp_vec_push(SleqpVec* vec,
                                         int idx,
                                         double value)

  # Sparse matrices
  SLEQP_RETCODE sleqp_mat_create(SleqpMat** matrix,
                                           int num_rows,
                                           int num_cols,
                                           int nnz_max)

  SLEQP_RETCODE sleqp_mat_reserve(SleqpMat* matrix,
                                            int nnz)

  SLEQP_RETCODE sleqp_mat_resize(SleqpMat* matrix,
                                           int num_rows,
                                           int num_cols)

  int sleqp_mat_num_cols(SleqpMat* matrix)

  int sleqp_mat_num_rows(SleqpMat* matrix)

  int sleqp_mat_nnz(SleqpMat* matrix)

  int sleqp_mat_nnz_max(SleqpMat* matrix)

  double* sleqp_mat_data(SleqpMat* matrix)

  int* sleqp_mat_cols(SleqpMat* matrix)

  int* sleqp_mat_rows(SleqpMat* matrix)

  SLEQP_RETCODE sleqp_mat_push(SleqpMat* matrix,
                                         int row,
                                         int col,
                                         double value)

  SLEQP_RETCODE sleqp_mat_push_col(SleqpMat* matrix,
                                                int col)

  SLEQP_RETCODE sleqp_mat_capture(SleqpMat* matrix)

  SLEQP_RETCODE sleqp_mat_release(SleqpMat** matrix)

  # Functions
  ctypedef SLEQP_RETCODE (*SLEQP_FUNC_SET)(SleqpFunc* func,
                                           SleqpVec* x,
                                           SLEQP_VALUE_REASON reason,
                                           bool* reject,
                                           void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_FUNC_NONZEROS)(SleqpFunc* func,
                                                int* obj_grad_nnz,
                                                int* cons_val_nnz,
                                                int* cons_jac_nnz,
                                                int* hess_prod_nnz,
                                                void* func_data);

  ctypedef SLEQP_RETCODE (*SLEQP_FUNC_OBJ_VAL)(SleqpFunc* func,
                                               double* func_val,
                                               void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_FUNC_OBJ_GRAD)(SleqpFunc* func,
                                                SleqpVec* func_grad,
                                                void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_FUNC_CONS_VAL)(SleqpFunc* func,
                                                SleqpVec* cons_val,
                                                void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_FUNC_CONS_JAC)(SleqpFunc* func,
                                                SleqpMat* cons_jac,
                                                void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_FUNC_HESS_PROD)(SleqpFunc* func,
                                                 const SleqpVec* direction,
                                                 const SleqpVec* cons_duals,
                                                 SleqpVec* product,
                                                 void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_FUNC_FREE)(void* func_data)

  ctypedef struct SleqpFuncCallbacks:
    SLEQP_FUNC_SET       set_value
    SLEQP_FUNC_NONZEROS  nonzeros
    SLEQP_FUNC_OBJ_VAL   obj_val
    SLEQP_FUNC_OBJ_GRAD  obj_grad
    SLEQP_FUNC_CONS_VAL  cons_val
    SLEQP_FUNC_CONS_JAC  cons_jac
    SLEQP_FUNC_HESS_PROD hess_prod
    SLEQP_FUNC_FREE      func_free

  SLEQP_RETCODE sleqp_func_create(SleqpFunc** fstar,
                                  SleqpFuncCallbacks* callbacks,
                                  int num_variables,
                                  int num_constraints,
                                  void* func_data)

  SLEQP_RETCODE sleqp_func_set_callbacks(SleqpFunc* func,
                                         SleqpFuncCallbacks* callbacks)

  int sleqp_func_num_vars(SleqpFunc* func)

  int sleqp_func_num_cons(SleqpFunc* func)

  SLEQP_RETCODE sleqp_func_capture(SleqpFunc* func)

  SLEQP_RETCODE sleqp_func_release(SleqpFunc** fstar)

  SleqpHessStruct* sleqp_func_hess_struct(SleqpFunc* func)

  # Hessian struct

  int sleqp_hess_struct_num_blocks(SleqpHessStruct* hess_struct)

  SLEQP_RETCODE sleqp_hess_struct_block_range(SleqpHessStruct* hess_struct,
                                              int block,
                                              int* begin,
                                              int* end)

  SLEQP_RETCODE sleqp_hess_struct_push_block(SleqpHessStruct* hess_struct,
                                             int end)

  SLEQP_RETCODE sleqp_hess_struct_clear(SleqpHessStruct* hess_struct)

  SLEQP_RETCODE sleqp_hess_struct_lin_range(SleqpHessStruct* hess_struct,
                                            int* begin,
                                            int* end)

  SLEQP_RETCODE sleqp_hess_struct_fprintf(SleqpHessStruct* hess_struct,
                                          libc.stdio.FILE* output)

  SLEQP_RETCODE sleqp_hess_struct_capture(SleqpHessStruct* hess_struct)

  SLEQP_RETCODE sleqp_hess_struct_release(SleqpHessStruct** star)

  # Iterate

  SLEQP_RETCODE sleqp_iterate_capture(SleqpIterate* iterate)

  SleqpVec* sleqp_iterate_primal(SleqpIterate* iterate)

  double sleqp_iterate_obj_val(SleqpIterate* iterate)

  SLEQP_RETCODE sleqp_iterate_set_obj_val(SleqpIterate* iterate,
                                          double value)

  SleqpVec* sleqp_iterate_obj_grad(SleqpIterate* iterate)

  SleqpVec* sleqp_iterate_cons_val(SleqpIterate* iterate)

  SleqpMat* sleqp_iterate_cons_jac(SleqpIterate* iterate)

  SleqpWorkingSet* sleqp_iterate_working_set(SleqpIterate* iterate)

  SleqpVec* sleqp_iterate_cons_dual(SleqpIterate* iterate)

  SleqpVec* sleqp_iterate_vars_dual(SleqpIterate* iterate)

  SLEQP_RETCODE sleqp_iterate_release(SleqpIterate** star)

  # Working set

  SLEQP_RETCODE sleqp_working_set_capture(SleqpWorkingSet* working_set)

  int sleqp_working_set_num_active_vars(const SleqpWorkingSet* working_set)

  int sleqp_working_set_num_active_cons(const SleqpWorkingSet* working_set)

  int sleqp_working_set_size(const SleqpWorkingSet* working_set)

  SleqpProblem* sleqp_working_set_problem(const SleqpWorkingSet* working_set)

  SLEQP_ACTIVE_STATE sleqp_working_set_var_state(const SleqpWorkingSet* working_set,
                                                 int index)

  SLEQP_ACTIVE_STATE sleqp_working_set_cons_state(const SleqpWorkingSet* working_set,
                                                  int index)

  SLEQP_RETCODE sleqp_working_set_release(SleqpWorkingSet** star)


  # LSQ

  ctypedef SLEQP_RETCODE (*SLEQP_LSQ_NONZEROS)(SleqpFunc* func,
                                               int* residual_nnz,
                                               int* jac_fwd_nnz,
                                               int* jac_adj_nnz,
                                               int* cons_val_nnz,
                                               int* cons_jac_nnz,
                                               void* func_data);

  ctypedef SLEQP_RETCODE (*SLEQP_LSQ_RESIDUALS)(SleqpFunc* func,
                                                SleqpVec* residual,
                                                void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_LSQ_JAC_FORWARD)(SleqpFunc* func,
                                                  SleqpVec* forward_direction,
                                                  SleqpVec* product,
                                                  void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_LSQ_JAC_ADJOINT)(SleqpFunc* func,
                                                  SleqpVec* adjoint_direction,
                                                  SleqpVec* product,
                                                  void* func_data)

  ctypedef struct SleqpLSQCallbacks:
    SLEQP_FUNC_SET        set_value,
    SLEQP_LSQ_NONZEROS    lsq_nonzeros
    SLEQP_LSQ_RESIDUALS   lsq_residuals,
    SLEQP_LSQ_JAC_FORWARD lsq_jac_forward,
    SLEQP_LSQ_JAC_ADJOINT lsq_jac_adjoint,
    SLEQP_FUNC_CONS_VAL   cons_val,
    SLEQP_FUNC_CONS_JAC   cons_jac,
    SLEQP_FUNC_FREE       func_free

  SLEQP_RETCODE sleqp_lsq_func_create(SleqpFunc** fstar,
                                      SleqpLSQCallbacks* callbacks,
                                      int num_variables,
                                      int num_constraints,
                                      int num_residuals,
                                      double levenberg_marquardt,
                                      SleqpSettings* settings,
                                      void* func_data)

  SLEQP_RETCODE sleqp_lsq_func_set_callbacks(SleqpFunc* func,
                                             SleqpLSQCallbacks* callbacks)

  # Dynamic

  ctypedef SLEQP_RETCODE (*SLEQP_DYN_FUNC_EVAL)(SleqpFunc* func,
                                                double* obj_val,
                                                SleqpVec* cons_val,
                                                double* error,
                                                void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_DYN_FUNC_SET_ERROR_BOUND)(SleqpFunc* func,
                                                           double error_bound,
                                                           void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_DYN_FUNC_SET_OBJ_WEIGHT)(SleqpFunc* func,
                                                          double obj_weight,
                                                          void* func_data)

  ctypedef SLEQP_RETCODE (*SLEQP_DYN_FUNC_SET_CONS_WEIGHTS)(SleqpFunc* func,
                                                            const double* cons_weights,
                                                            void* func_data)

  ctypedef struct SleqpDynFuncCallbacks:
    SLEQP_FUNC_SET set_value,
    SLEQP_FUNC_NONZEROS nonzeros,
    SLEQP_DYN_FUNC_SET_ERROR_BOUND set_error_bound,
    SLEQP_DYN_FUNC_SET_OBJ_WEIGHT set_obj_weight,
    SLEQP_DYN_FUNC_SET_CONS_WEIGHTS set_cons_weights,
    SLEQP_DYN_FUNC_EVAL eval,
    SLEQP_FUNC_OBJ_GRAD obj_grad,
    SLEQP_FUNC_CONS_JAC cons_jac,
    SLEQP_FUNC_HESS_PROD hess_prod,
    SLEQP_FUNC_FREE func_free

  SLEQP_RETCODE sleqp_dyn_func_create(SleqpFunc** fstar,
                                      SleqpDynFuncCallbacks* callbacks,
                                      int num_variables,
                                      int num_constraints,
                                      void* func_data)

  SLEQP_RETCODE sleqp_dyn_func_set_callbacks(SleqpFunc* func,
                                             SleqpDynFuncCallbacks* callbacks)

  # Scaling

  SLEQP_RETCODE sleqp_scaling_create(SleqpScaling** scaling,
                                     int num_variables,
                                     int num_constraints)

  int sleqp_scaling_num_vars(SleqpScaling* scaling)
  int sleqp_scaling_num_cons(SleqpScaling* scaling)

  int sleqp_scaling_obj_weight(SleqpScaling* scaling)

  SLEQP_RETCODE sleqp_scaling_set_obj_weight(SleqpScaling* scaling,
                                             int weight)

  SLEQP_RETCODE sleqp_scaling_set_obj_weight_from_nominal(SleqpScaling* scaling,
                                                          double nominal_value)

  SLEQP_RETCODE sleqp_scaling_set_var_weight(SleqpScaling* scaling,
                                             int index,
                                             int weight)

  SLEQP_RETCODE sleqp_scaling_set_var_weight_from_nominal(SleqpScaling* scaling,
                                                          int index,
                                                          double nominal_value)

  SLEQP_RETCODE sleqp_scaling_set_var_weights_from_nominal(SleqpScaling* scaling,
                                                           double* nominal_values)

  SLEQP_RETCODE sleqp_scaling_set_cons_weight(SleqpScaling* scaling,
                                              int index,
                                              int weight)

  int* sleqp_scaling_var_weights(SleqpScaling* scaling)

  int* sleqp_scaling_cons_weights(SleqpScaling* scaling)

  SLEQP_RETCODE sleqp_scaling_set_cons_weights_from_nominal(SleqpScaling* scaling,
                                                            double* nominal_values)

  SLEQP_RETCODE sleqp_scaling_set_cons_weight_from_nominal(SleqpScaling* scaling,
                                                           int index,
                                                           double nominal_value);

  SLEQP_RETCODE sleqp_obj_scaling_from_grad(SleqpScaling* scaling,
                                            SleqpVec* gradient,
                                            double eps)

  SLEQP_RETCODE sleqp_scaling_from_cons_jac(SleqpScaling* scaling,
                                            SleqpMat* cons_jac,
                                            double eps)

  SLEQP_RETCODE sleqp_scaling_release(SleqpScaling** scaling)

  # Solver

  SLEQP_RETCODE sleqp_solver_create(SleqpSolver** star,
                                    SleqpProblem* problem,
                                    SleqpVec* x,
                                    SleqpScaling* scaling)

  const char* sleqp_solver_info(const SleqpSolver* solver)

  SLEQP_RETCODE sleqp_solver_solve(SleqpSolver* solver,
                                   int max_num_iterations,
                                   double time_limit) nogil

  SLEQP_RETCODE sleqp_solver_real_state(const SleqpSolver* solver,
                                        SLEQP_SOLVER_STATE_REAL state,
                                        double* value)

  SLEQP_RETCODE sleqp_solver_int_state(const SleqpSolver* solver,
                                       SLEQP_SOLVER_STATE_INT state,
                                       int* value)

  SLEQP_RETCODE sleqp_solver_vec_state(const SleqpSolver* solver,
                                       SLEQP_SOLVER_STATE_VEC value,
                                       SleqpVec* result)

  SLEQP_STATUS sleqp_solver_status(SleqpSolver* solver)

  SLEQP_RETCODE sleqp_solver_abort(SleqpSolver* solver)

  int sleqp_solver_iterations(SleqpSolver* solver)

  double sleqp_solver_elapsed_seconds(SleqpSolver* solver)

  SLEQP_RETCODE sleqp_solver_add_callback(SleqpSolver* solver,
                                          SLEQP_SOLVER_EVENT solver_event,
                                          void* callback_func,
                                          void* callback_data)

  SLEQP_RETCODE sleqp_solver_remove_callback(SleqpSolver* solver,
                                             SLEQP_SOLVER_EVENT solver_event,
                                             void* callback_func,
                                             void* callback_data)

  SLEQP_RETCODE sleqp_solver_solution(SleqpSolver* solver,
                                      SleqpIterate** iterate)

  SLEQP_RETCODE sleqp_solver_violated_constraints(SleqpSolver* solver,
                                                  SleqpIterate* iterate,
                                                  int* violated_constraints,
                                                  int* num_violated_constraints)

  SLEQP_RETCODE sleqp_solver_release(SleqpSolver** star)

  # Problem

  SLEQP_RETCODE sleqp_problem_create_simple(SleqpProblem** star,
                                            SleqpFunc* func,
                                            const SleqpVec* var_lb,
                                            const SleqpVec* var_ub,
                                            const SleqpVec* cons_lb,
                                            const SleqpVec* cons_ub,
                                            SleqpSettings* settings)

  SLEQP_RETCODE sleqp_problem_create(SleqpProblem** star,
                                     SleqpFunc* func,
                                     const SleqpVec* var_lb,
                                     const SleqpVec* var_ub,
                                     const SleqpVec* genereal_lb,
                                     const SleqpVec* genereal_ub,
                                     const SleqpMat* linear_coeffs,
                                     const SleqpVec* linear_lb,
                                     const SleqpVec* linear_ub,
                                     SleqpSettings* settings)

  int sleqp_problem_num_cons(SleqpProblem* problem)

  int sleqp_problem_num_vars(SleqpProblem* problem)

  SleqpVec* sleqp_problem_vars_lb(SleqpProblem* problem)

  SleqpVec* sleqp_problem_vars_ub(SleqpProblem* problem)

  SleqpVec* sleqp_problem_cons_lb(SleqpProblem* problem)

  SleqpVec* sleqp_problem_cons_ub(SleqpProblem* problem)

  SLEQP_RETCODE sleqp_problem_release(SleqpProblem** star)

  # Settings

  SLEQP_RETCODE sleqp_settings_create(SleqpSettings** star)

  SLEQP_RETCODE sleqp_settings_read_file(SleqpSettings* settings,
                                         const char* settings_filename)

  int sleqp_settings_enum_value(const SleqpSettings* options,
                                SLEQP_SETTINGS_ENUM option)

  SLEQP_RETCODE sleqp_settings_set_enum_value(SleqpSettings* options,
                                              SLEQP_SETTINGS_ENUM option,
                                              int value)

  int sleqp_settings_int_value(const SleqpSettings* options,
                               SLEQP_SETTINGS_INT option)

  SLEQP_RETCODE sleqp_settings_set_int_value(SleqpSettings* options,
                                             SLEQP_SETTINGS_INT option,
                                             int value)


  bool sleqp_settings_bool_value(const SleqpSettings* options,
                                 SLEQP_SETTINGS_BOOL option)

  SLEQP_RETCODE sleqp_settings_set_bool_value(SleqpSettings* options,
                                              SLEQP_SETTINGS_BOOL option,
                                              bool value)

  double sleqp_settings_real_value(const SleqpSettings* settings,
                                   SLEQP_SETTINGS_REAL param)

  SLEQP_RETCODE sleqp_settings_set_real_value(SleqpSettings* settings,
                                              SLEQP_SETTINGS_REAL param,
                                              double value)


  SLEQP_RETCODE sleqp_settings_release(SleqpSettings** star)

  # Logging

  void sleqp_log_set_level(SLEQP_LOG_LEVEL value)

  ctypedef void (*SLEQP_LOG_HANDLER)(SLEQP_LOG_LEVEL level,
                                     libc.time.time_t time,
                                     const char* message)

  void sleqp_log_set_handler(SLEQP_LOG_HANDLER handler)

  SLEQP_LOG_LEVEL sleqp_log_level()

  # Numerics
  double sleqp_infinity()



#cython: language_level=3

import numpy as np
cimport numpy as np

import enum
import traceback
import typing
import collections.abc
import weakref

import scipy.sparse

cimport csleqp

# This has to be the first import
include "types.pxi"

include "gil.pxi"
include "call.pxi"
include "callback.pxi"
include "cmp.pxi"
include "dyn.pxi"
include "sparse.pxi"
include "sparse_matrix.pxi"
include "func.pxi"
include "hess_struct.pxi"
include "iterate.pxi"
include "log.pxi"
include "lsq.pxi"
include "settings.pxi"
include "problem.pxi"
include "scale.pxi"
include "solver.pxi"
include "version.pxi"
include "working_set.pxi"



import numpy as np

FD_METHODS = ('2-point', '3-point', 'cs')

EPS = np.finfo(float).eps

eps_twopoint = EPS**(1/2)
eps_threepoint = EPS**(1/3)

def perturbation(x, eps):
  return eps * np.maximum(np.abs(x), 1.)


def findiff_twopoint(f):

  def evaluate(x0, args=(), f0=None):
    x0 = np.atleast_1d(x0)
    h = perturbation(x0, eps_twopoint)

    if f0 is None:
      f0 = f(x0, *args)

    n = x0.size
    m = f0.size

    xd = np.copy(x0)

    df = np.atleast_2d(np.empty((*f0.shape, n)))

    for j in range(n):
      xd[j] += h[j]

      fd = f(xd, *args)

      df[:, j] = (fd - f0) / h[j]

      xd[j] = x0[j]

    return df

  return evaluate


def findiff_threepoint(f):

  def evaluate(x0, args=(), f0=None):
    x0 = np.atleast_1d(x0)
    h = perturbation(x0, eps_threepoint)

    if f0 is None:
      f0 = f(x0, *args)

    n = x0.size
    m = f0.size

    xd = np.copy(x0)

    df = np.atleast_2d(np.empty((*f0.shape, n)))

    for j in range(n):
      xd[j] += h[j]

      fpos = f(xd, *args)

      xd[j] = x0[j] - h[j]

      fneg = f(xd, *args)

      xd[j] = x0[j]

      df[:, j] = (fpos - fneg) / (2 * h[j])

    return df

  return evaluate


def findiff_cs(f):
  def evaluate(x0, args=(), f0=None):
    x0 = np.atleast_1d(x0)
    h = perturbation(x0, eps_twopoint)

    n = x0.size
    m = f0.size

    if f0 is None:
      f0 = f(x0, *args)

    df = np.atleast_2d(np.empty((*f0.shape, n)))

    xd = x0.astype('complex')

    im = 1.j

    for j in range(n):
      xd[j] += h[j]*im

      df[:, j] = np.imag(f(xd)) / h[j]

      xd[j] = x0[j]

    return df

  return evaluate


def derivative(jac):
  def evaluate(x0, args=(), f0=None):
    return np.atleast_1d(jac(x0, *args))

  return evaluate

def create_derivative(fun, jac):
  def func(x, *args):
    return np.atleast_1d(fun(x, *args))

  if callable(jac):
    return derivative(jac)
  elif jac == '2-point':
    return findiff_twopoint(func)
  elif jac == '3-point':
    return findiff_threepoint(func)
  elif jac == 'cs':
    return findiff_cs(func)
  elif jac is None:
    return findiff_twopoint(func)

  raise ValueError("Invalid Jacobian: %s" % jac)



#cython: language_level=3

cdef class WorkingSet:
  """
  The working set associated with an iterate, containing
  the states of all variables and constraints.
  """
  cdef csleqp.SleqpWorkingSet* working_set

  cdef dict __dict__

  cdef _release(self):
    if self.working_set:
      csleqp_call(csleqp.sleqp_working_set_release(&self.working_set))

  cdef _set_working_set(self, csleqp.SleqpWorkingSet* working_set):
    self._release()

    csleqp_call(csleqp.sleqp_working_set_capture(working_set))

    self.working_set = working_set

  @property
  def num_vars(self) -> int:
    """
    The total number of (active or inactive) variables in the working set
    """
    cdef csleqp.SleqpProblem* problem

    problem = csleqp.sleqp_working_set_problem(self.working_set)

    return csleqp.sleqp_problem_num_vars(problem)

  @property
  def num_cons(self) -> int:
    """
    The total number of (active or inactive) constraints in the working set
    """
    cdef csleqp.SleqpProblem* problem

    problem = csleqp.sleqp_working_set_problem(self.working_set)

    return csleqp.sleqp_problem_num_cons(problem)

  @property
  def num_active_vars(self) -> int:
    """
    The total number of active (lower or upper or both)
    variables in the working set
    """
    assert self.working_set

    return csleqp.sleqp_working_set_num_active_vars(self.working_set)

  @property
  def num_active_cons(self) -> int:
    """
    The total number of active (lower or upper or both)
    constraints in the working set
    """
    assert self.working_set

    return csleqp.sleqp_working_set_num_active_cons(self.working_set)

  def var_state(self, int i) -> ActiveState:
    """
    State of variable i
    """
    assert self.working_set

    state = csleqp.sleqp_working_set_var_state(self.working_set,
                                               i)

    return ActiveState(state)

  def cons_state(self, int i) -> ActiveState:
    """
    State of constraint i
    """
    assert self.working_set

    state = csleqp.sleqp_working_set_cons_state(self.working_set,
                                                i)

    return ActiveState(state)

  def __str__(self):
    str_val = "Active set, variables: {0}, constraints: {1}\n".format(self.num_vars,
                                                                      self.num_cons)

    str_val += "\n".join(("State of variable {0}: {1}".format(j, str(self.var_state(j)))
                          for j in range(self.num_vars)))

    str_val += "\n"

    str_val += "\n".join(("State of constraint {0}: {1}".format(i, str(self.cons_state(i)))
                          for i in range(self.num_cons)))

    return str_val

  @property
  def size(self) -> int:
    """
    Size of the working set, i.e., number of active variables
    and active constraints
    """
    assert self.working_set

    return csleqp.sleqp_working_set_size(self.working_set)

  def __dealloc__(self):
    self._release()



import numpy as np

import sleqp


def create_variable_bounds(num_variables, bounds):
  inf = sleqp.inf()

  var_lb = np.full((num_variables,), -inf)
  var_ub = np.full((num_variables,), inf)

  if bounds is None:
    return (var_lb, var_ub)

  try:
    var_lb = np.array(bounds.lb)
    var_ub = np.array(bounds.ub)
    return (var_lb, var_ub)
  except AttributeError:
    pass

  for i, (lb, ub) in enumerate(bounds):
    if lb is not None:
      var_lb[i] = lb
    if ub is not None:
      var_ub[i] = ub

  return (var_lb, var_ub)


def create_constraint_bounds(constraints):

  inf = sleqp.inf()
  cons_lb = []
  cons_ub = []

  for i, constraint in enumerate(constraints):

    try:
      cons_lb.append(constraint.lb)
      cons_ub.append(constraint.ub)
      continue
    except AttributeError:
      pass

    if constraint['type'] == 'ineq':
      cons_lb.append(np.array([0.]))
      cons_ub.append(np.array([inf]))
    else:
      assert constraint['type'] == 'eq'
      cons_lb.append(np.array([0.]))
      cons_ub.append(np.array([0.]))

  if cons_lb:
    cons_lb = np.hstack(cons_lb)
  else:
    cons_lb = np.array([])

  if cons_ub:
    cons_ub = np.hstack(cons_ub)
  else:
    cons_ub = np.array([])

  assert cons_lb.shape == cons_ub.shape

  assert (cons_lb <= cons_ub).all()

  return (cons_lb, cons_ub)



from dataclasses import dataclass

import numpy as np

from scipy.optimize import (LinearConstraint,
                            NonlinearConstraint)

from sleqp._bounds import create_constraint_bounds
from sleqp._cons_func import create_constraint_func

from sleqp._linear import (create_linear_constraints, LinearConstraints)


class GeneralConstraints:
  def __init__(self,
               cons_lb=np.zeros((0,)),
               cons_func=None,
               cons_ub=np.zeros((0,))):
    self.lb = cons_lb
    self.func = cons_func
    self.ub = cons_ub


class Constraints:
  def __init__(self,
               general=GeneralConstraints(),
               linear=LinearConstraints()):
    self.general = general
    self.linear = linear


def split_linear_constraints(constraints):
  general = []
  linear = []

  for constraint in constraints:
    if isinstance(constraint, LinearConstraint):
      linear.append(constraint)
    else:
      general.append(constraint)

  return (general, linear)


def create_general_constraints(num_variables, constraints):
  (cons_lb, cons_ub) = create_constraint_bounds(constraints)

  cons_func = create_constraint_func(num_variables, constraints)

  return GeneralConstraints(cons_lb, cons_func, cons_ub)


def create_constraints(num_variables, constraints):

  if constraints is None:
    return Constraints()

  if not (isinstance(constraints, list) or
          isinstance(constraints, tuple)):
    constraints = [constraints]

  (general, linear) = split_linear_constraints(constraints)

  general = create_general_constraints(num_variables, general)
  linear = create_linear_constraints(linear)

  return Constraints(general=general, linear=linear)



from setuptools import setup, find_packages
from setuptools.extension import Extension

import os
import subprocess

import numpy

project_description = 'Active set-based NLP solver'
long_description = 'This package provides SLEQP, an open-source solver for large-scale nonlinear continuous optimization. SLEQP is available as a callable library with interfaces to C, Python, AMPL, and MATLAB / Octave. It uses an active set method based on a combination of successive linear programming and equality constrained quadratic programming.'
project_name = 'sleqp'
project_version = '1.0.1'
license_file = 'LICENSE.txt'
maintainer_name = 'Christoph Hansknecht'
maintainer_email = 'c.hansknecht@tu-braunschweig.de'

local_env = os.getenv('SLEQP_LOCAL_BUILD')
local_build = (local_env is not None)

ldflags = []
cflags = []
lib_dirs = []
include_dirs = []


def pkgconfig(name):

    def pkgconfig_get(flags):
        res = subprocess.run(['pkg-config', flags, name],
                             stdout=subprocess.PIPE,
                             check=True)

        return res.stdout.decode().strip().split()

    cflags = pkgconfig_get('--cflags-only-other')

    includes = pkgconfig_get('--cflags-only-I')

    libs = pkgconfig_get('--libs-only-l')

    libdirs = pkgconfig_get('--libs-only-L')

    return {'cflags': cflags,
            'includes': includes,
            'libs': libs,
            'libdirs': libdirs}


if not(local_build):
    pkgconfig_args = pkgconfig(project_name)

    ldflags += pkgconfig_args['libs']
    cflags += pkgconfig_args['cflags']
    lib_dirs += pkgconfig_args['libdirs']
    include_dirs += pkgconfig_args['includes']


extensions = [
    Extension("sleqp.sleqp",
              ['src/extension/sleqp.pyx'],
              include_dirs=[numpy.get_include()] + include_dirs,
              libraries=[project_name],
              extra_link_args=ldflags,
              library_dirs=lib_dirs,
              extra_compile_args=cflags
              )]

requirements = []

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(name=project_name,
      author=maintainer_name,
      author_email=maintainer_email,
      description=(project_description),
      license_file=license_file,
      long_description=long_description,
      version=project_version,
      ext_modules=extensions,
      packages=find_packages(where="src", exclude=["tests"]),
      package_dir={"": "src"},
      install_requires=requirements,
      setup_requires=['wheel'],
      classifiers=[
        'Programming Language :: Cython',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics'
      ],
      test_suite='tests')



#cython: language_level=3

from enum import Enum


class _PropType(Enum):
  Bool = 0
  Integer = 1
  Enumerated = 2
  Real = 3


class _NullConverter:
  def from_py(self, val):
    return val

  def to_py(self, val):
    return val


class _EnumConverter:
  def __init__(self, enum_type):
    self.enum_type = enum_type

  def from_py(self, val):
    try:
      return val.value
    except AttributeError:
      pass

    try:
      return self.enum_type[val].value
    except KeyError:
      raise AttributeError('Invalid attribute value \'{0}\''.format(val))

  def to_py(self, val):
    return self.enum_type(val)


class _Prop:
  def __init__(self, value, prop_type, converter):
    self.value = value
    self.prop_type = prop_type
    self.converter = converter

  def to_py(self, val):
    return self.converter.to_py(val)

  def from_py(self, val):
    return self.converter.from_py(val)

  @staticmethod
  def boolean(value):
    return _Prop(value, _PropType.Bool, _NullConverter())

  @staticmethod
  def integer(value):
    return _Prop(value, _PropType.Integer, _NullConverter())

  @staticmethod
  def enumerated(value, enum_type):
    return _Prop(value, _PropType.Enumerated, _EnumConverter(enum_type))

  @staticmethod
  def real(value):
    return _Prop(value, _PropType.Real, _NullConverter())


cdef dict opt_prop_map = {
  # Boolean properties
  'perform_newton_step':   _Prop.boolean(csleqp.SLEQP_SETTINGS_BOOL_PERFORM_NEWTON_STEP),
  'global_penalty_resets': _Prop.boolean(csleqp.SLEQP_SETTINGS_BOOL_GLOBAL_PENALTY_RESETS),
  'perform_soc':           _Prop.boolean(csleqp.SLEQP_SETTINGS_BOOL_PERFORM_SOC),
  'use_quadratic_model':   _Prop.boolean(csleqp.SLEQP_SETTINGS_BOOL_USE_QUADRATIC_MODEL),
  'enable_restoration':    _Prop.boolean(csleqp.SLEQP_SETTINGS_BOOL_ENABLE_RESTORATION_PHASE),
  'enable_preprocessor':   _Prop.boolean(csleqp.SLEQP_SETTINGS_BOOL_ENABLE_PREPROCESSOR),
  'lp_resolves':           _Prop.boolean(csleqp.SLEQP_SETTINGS_BOOL_LP_RESOLVES),

  # Integer properties
  'num_quasi_newton_iterates': _Prop.integer(csleqp.SLEQP_SETTINGS_INT_NUM_QUASI_NEWTON_ITERATES),
  'max_newton_iterations':     _Prop.integer(csleqp.SLEQP_SETTINGS_INT_MAX_NEWTON_ITERATIONS),
  'num_threads':               _Prop.integer(csleqp.SLEQP_SETTINGS_INT_NUM_THREADS),

  # SLEQP_SETTINGS_INT_FLOAT_WARNING_FLAGS,
  # SLEQP_SETTINGS_INT_FLOAT_ERROR_FLAGS,

  # Enumerated properties
  'deriv_check':          _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_DERIV_CHECK, DerivCheck),
  'hessian_eval':         _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_HESS_EVAL, HessianEval),
  'dual_estimation_type': _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_DUAL_ESTIMATION_TYPE, DualEstimationType),
  'initial_tr_choice':    _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_INITIAL_TR_CHOICE, InitialTRChoice),
  'bfgs_sizing':          _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_BFGS_SIZING, Sizing),
  'tr_solver':            _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_TR_SOLVER, TRSolver),
  'polishing_type':       _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_POLISHING_TYPE, PolishingType),
  'step_rule':            _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_STEP_RULE, StepRule),
  'linesearch':           _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_LINESEARCH, LineSearch),
  'parametric_cauchy':    _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_PARAMETRIC_CAUCHY, ParametricCauchy),
  'aug_jac_method':       _Prop.enumerated(csleqp.SLEQP_SETTINGS_ENUM_AUG_JAC_METHOD, AugJacMethod),

  'zero_eps':           _Prop.real(csleqp.SLEQP_SETTINGS_REAL_ZERO_EPS),
  'eps':                _Prop.real(csleqp.SLEQP_SETTINGS_REAL_EPS),
  'deriv_perturbation': _Prop.real(csleqp.SLEQP_SETTINGS_REAL_DERIV_PERTURBATION),
  'deriv_tol':          _Prop.real(csleqp.SLEQP_SETTINGS_REAL_DERIV_TOL),
  'cauchy_tau':         _Prop.real(csleqp.SLEQP_SETTINGS_REAL_CAUCHY_TAU),
  'cauchy_eta':         _Prop.real(csleqp.SLEQP_SETTINGS_REAL_CAUCHY_ETA),
  'linesearch_tau':     _Prop.real(csleqp.SLEQP_SETTINGS_REAL_LINESEARCH_TAU),
  'linesearch_eta':     _Prop.real(csleqp.SLEQP_SETTINGS_REAL_LINESEARCH_ETA),
  'linesearch_cutoff':  _Prop.real(csleqp.SLEQP_SETTINGS_REAL_LINESEARCH_CUTOFF),
  'feas_tol':           _Prop.real(csleqp.SLEQP_SETTINGS_REAL_FEAS_TOL),
  'slack_tol':          _Prop.real(csleqp.SLEQP_SETTINGS_REAL_SLACK_TOL),
  'stat_tol':           _Prop.real(csleqp.SLEQP_SETTINGS_REAL_STAT_TOL),
  'accepted_reduction': _Prop.real(csleqp.SLEQP_SETTINGS_REAL_ACCEPTED_REDUCTION),
  'deadpoint_bound':    _Prop.real(csleqp.SLEQP_SETTINGS_REAL_DEADPOINT_BOUND)
}

cdef class Settings:
  cdef csleqp.SleqpSettings* settings

  def __cinit__(self):
    csleqp_call(csleqp.sleqp_settings_create(&self.settings))

  def __init__(self, **values):
    for key, value in values.items():
      setattr(self, key, value)

  def props(self) -> dict:
    return {key: getattr(self, key) for key in opt_prop_map}

  def __str__(self) -> str:
    return 'Settings: {0}'.format(self.props())

  def __repr__(self) -> str:
    return 'Settings({0})'.format(repr(self.props()))

  cpdef read(self, str filename):
    csleqp_call(csleqp.sleqp_settings_read_file(self.settings,
                                                filename))

  cdef _prop(self, name):
    prop = opt_prop_map.get(name)
    if prop is None:
      raise AttributeError('Settings has no attribute \'{0}\''.format(name))
    return prop

  def __getattr__(self, name):
    prop = self._prop(name)
    prop_val = None

    if prop.prop_type == _PropType.Integer:
      prop_val = csleqp.sleqp_settings_int_value(self.settings,
                                                 prop.value)
    elif prop.prop_type == _PropType.Real:
      prop_val = csleqp.sleqp_settings_real_value(self.settings,
                                                  prop.value)
    elif prop.prop_type == _PropType.Enumerated:
      prop_val = csleqp.sleqp_settings_enum_value(self.settings,
                                                  prop.value)
    else:
      assert prop.prop_type == _PropType.Bool

      prop_val = csleqp.sleqp_settings_bool_value(self.settings,
                                                  prop.value)

    return prop.to_py(prop_val)

  def __setattr__(self, name, value):
    prop = self._prop(name)
    prop_val = prop.from_py(value)

    if prop.prop_type == _PropType.Integer:
      csleqp_call(csleqp.sleqp_settings_set_int_value(self.settings,
                                                      prop.value,
                                                      prop_val))
    elif prop.prop_type == _PropType.Real:
      csleqp_call(csleqp.sleqp_settings_set_real_value(self.settings,
                                                       prop.value,
                                                       prop_val))
    elif prop.prop_type == _PropType.Enumerated:
      csleqp_call(csleqp.sleqp_settings_set_enum_value(self.settings,
                                                       prop.value,
                                                       prop_val))
    else:
      assert prop.prop_type == _PropType.Bool

      csleqp_call(csleqp.sleqp_settings_set_bool_value(self.settings,
                                                       prop.value,
                                                       prop_val))



import logging

import numpy as np
from scipy.optimize import OptimizeResult

import sleqp

from sleqp._bounds import create_variable_bounds
from sleqp._cons import create_constraints
from sleqp._func import create_func
from sleqp._hessian import create_hessian


class _MinFunc:

  def __init__(self, objective, constraints, args, num_variables):
    self.x = np.zeros((num_variables,))
    self.objective = objective
    self.constraints = constraints
    self.args = args
    self.num_variables = num_variables

    self._obj_val = None
    self._obj_grad = None

  def set_value(self, v, reason):
    self.x[:] = v

    self.objective.set_value(self.x)

    if self.constraints:
      self.constraints.set_value(self.x)

    self._obj_val = None
    self._obj_grad = None

  def cons_vals(self):
    return self.constraints.val(self.args)

  def cons_jac(self):
    return self.constraints.jac(self.args)

  def obj_val(self):
    self._eval_obj_val()
    return self._obj_val

  def _eval_obj_val(self):
    if self._obj_val is None:
      self._obj_val = self.objective.val(self.args)

  def _eval_obj_grad(self):
    self._eval_obj_val()

    if self._obj_grad is None:
      self._obj_grad = self.objective.grad(self.args)

      if self._obj_grad.ndim == 2:
        self._obj_grad = self._obj_grad[0,:]

  def obj_grad(self):
    self._eval_obj_grad()
    return self._obj_grad

  def hess_prod(self, direction, cons_duals):
    self._eval_obj_grad()

    prod = self.objective.hess_prod(direction,
                                    self.args)

    if self.constraints:
      cons_prod = self.constraints.hess_prod(direction,
                                             cons_duals,
                                             self.args)
      prod += cons_prod

    return prod


def _create_result(solver):
  solution = solver.solution

  result = OptimizeResult()

  result["mult_g"] = solution.cons_dual
  result["mult_x"] = solution.vars_dual
  result["x"] = solution.primal
  result["success"] = (solver.status == sleqp.Status.Optimal)
  result["status"] = solver.status.value
  result["message"] = solver.status.desc

  result["fun"] = solution.obj_val
  result["jac"] = solution.obj_grad
  result["nit"] = solver.iterations

  result["maxcv"] = solver.states[sleqp.SolverState.ScaledFeasRes]

  return result

def _add_solver_callback(solver, callback):

  def accepted_iterate(solver, iterate):
    abort = callback(iterate.primal)

    if abort is True:
      solver.abort()

  solver.add_callback(sleqp.SolverEvent.AcceptedIterate,
                      accepted_iterate)


_handler = None
_attached = False

def _attach_handler():
  global _handler, _attached

  if _handler is None:
    _handler = logging.StreamHandler()
    _handler.setLevel(logging.INFO)

  if not(_attached):
    sleqp.logger.addHandler(_handler)
    _attached = True


def _detach_handler():
  global _handler, _attached

  if _handler is None:
    return

  if _attached:
    sleqp.logger.removeHandler(_handler)
    _attached = False


class LogHelper:
  def __init__(self, verbose):
    self.verbose=verbose
    self.old_level=None

  def __enter__(self):
    if not(self.verbose):
      return

    # Set level to at least INFO
    if sleqp.logger.level < logging.INFO:
      self.old_level = sleqp.logger.level
      sleqp.logger.setLevel(logging.INFO)

    _attach_handler()

  def __exit__(self, exc_type, exc_value, exc_tb):
    if self.verbose:
      _detach_handler()

    # Restore old level
    if self.old_level is not None:
      sleqp.logger.setLevel(self.old_level)


def minimize(fun, x0, args=(), jac=None, hess=None, hessp=None, bounds=None, constraints=None, callback=None, **options):
  """
  A drop-in replacement for :func:`scipy.optimize.minimize`, minimizing a scalar function
  of one or more variables subjecting to constraints.

  :param fun:
      The objective function to be minimized.
      ``fun(x, *args) -> float``
      where ``x`` is an 1-D array with shape (n,) and ``args``
      is a tuple of the fixed parameters needed to completely
      specify the function.
  :type fun: callable
  :param x0:
        Initial guess. Array of real elements of size (n,),
        where 'n' is the number of independent variables.
  :type x0: :class:`numpy.ndarray`, shape (n,)
  :param args:
        Extra arguments passed to the objective function and its
        derivatives (`fun`, `jac` and `hess` functions).
  :type args: tuple, optional
  :param bounds:
        Bounds on variables. There are two ways to specify the bounds:
        1. Instance of :class:`scipy.optimize.Bounds` class.
        2. Sequence of ``(min, max)`` pairs for each element in `x`. ``None`` is used to specify no bound.
  :type bounds: sequence or :class:`scipy.optimize.Bounds`, optional
  :param callback:
        Called after each iteration. If callback returns True
        the algorithm execution is terminated. The signature is: ``callback(xk)``
        where ``xk`` is the current guess.
  :type callback: callable, optional

  :return: The optimization result
  :rtype: :class:`scipy.optimize.Optimizeresult`
  """
  if not isinstance(args, tuple):
    args = (args,)

  num_variables = len(x0)

  cons_lb = np.zeros((0,))
  cons_ub = np.zeros((0,))
  cons_func = None

  constraints = create_constraints(num_variables, constraints)

  objective = create_func(fun, jac, hess, hessp)

  verbose = options.pop('verbose', False)

  settings = sleqp.Settings(**options)

  if hessp is None and hess is None:
    settings.hessian_eval = sleqp.HessianEval.DampedBFGS

  initial_sol = np.array(x0)

  (var_lb, var_ub) = create_variable_bounds(num_variables, bounds)

  min_func = _MinFunc(objective,
                      constraints.general.func,
                      args,
                      num_variables)

  problem = sleqp.Problem(min_func,
                          var_lb,
                          var_ub,
                          constraints.general.lb,
                          constraints.general.ub,
                          settings,
                          linear_coeffs=constraints.linear.coeffs,
                          linear_lb=constraints.linear.lb,
                          linear_ub=constraints.linear.ub)

  solver = sleqp.Solver(problem,
                        initial_sol)

  if callback is not None:
    _add_solver_callback(solver, callback)

  with LogHelper(verbose):
    solver.solve()

  return _create_result(solver)



cdef bint release_gil = False

cpdef get_release_gil():
  """
  Returns whether or not the GIL is released during the solution process

  :rtype: bool
  """
  return release_gil

cpdef set_release_gil(bint value):
  """
  Advising the solver whether or not to release the GIL during
  the solution process (it will be reacquired during function / callback evaluations)

  :param value: whether or not to relase the GIL
  :type value: bool
  """
  global release_gil

  if value == release_gil:
    return

  release_gil = value
  update_log_handler()
  update_func_callbacks()
  update_dyn_func_callbacks()
  update_lsq_func_callbacks()
  update_solver_callbacks()



#cython: language_level=3

cdef csleqp.SLEQP_RETCODE sleqp_dyn_eval(csleqp.SleqpFunc* func,
                                         double* obj_val,
                                         csleqp.SleqpVec* cons_val,
                                         double* error,
                                         void* func_data):
  try:
    func_obj = (<object> func_data)

    result = func_obj.eval()

    assert result is not None, "eval() returned 'None'"

    num_cons = csleqp.sleqp_func_num_cons(func)

    obj_val[0] = result[0]
    error[0] = result[1]

    if num_cons > 0:
      csleqp_call(array_to_sleqp_sparse_vec(result[2], cons_val))

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

  return csleqp.SLEQP_OKAY

cdef csleqp.SLEQP_RETCODE sleqp_dyn_eval_nogil(csleqp.SleqpFunc* func,
                                               double* obj_val,
                                               csleqp.SleqpVec* cons_val,
                                               double* error,
                                               void* func_data) nogil:
  with gil:
    return sleqp_dyn_eval(func,
                          obj_val,
                          cons_val,
                          error,
                          func_data)

cdef csleqp.SLEQP_RETCODE sleqp_dyn_set_error_bound(csleqp.SleqpFunc* func,
                                                    double error_bound,
                                                    void* func_data):
  try:
    func_obj = (<object> func_data)

    func_obj.set_error_bound(error_bound)

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

cdef csleqp.SLEQP_RETCODE sleqp_dyn_set_error_bound_nogil(csleqp.SleqpFunc* func,
                                                          double error_bound,
                                                          void* func_data) nogil:
  with gil:
    return sleqp_dyn_set_error_bound(func,
                                     error_bound,
                                     func_data)

cdef csleqp.SLEQP_RETCODE sleqp_dyn_set_obj_weight(csleqp.SleqpFunc* func,
                                                   double obj_weight,
                                                   void* func_data):
  try:
    func_obj = (<object> func_data)

    func_obj.set_obj_weight(obj_weight)

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

cdef csleqp.SLEQP_RETCODE sleqp_dyn_set_obj_weight_nogil(csleqp.SleqpFunc* func,
                                                         double obj_weight,
                                                         void* func_data) nogil:
  with gil:
    return sleqp_dyn_set_obj_weight(func,
                                    obj_weight,
                                    func_data)

cdef csleqp.SLEQP_RETCODE sleqp_dyn_set_cons_weights(csleqp.SleqpFunc* func,
                                                     const double* cons_weights,
                                                     void* func_data):
  cdef double[:] cons_view
  try:
    func_obj = (<object> func_data)

    num_cons = csleqp.sleqp_func_num_cons(func)
    cons_view = _readonly(<double[:num_cons]> cons_weights)

    func_obj.set_cons_weights(np.copy(cons_view))

  except BaseException as exception:
    store_func_exc(func_obj, exception)
    return csleqp.SLEQP_ERROR

cdef csleqp.SLEQP_RETCODE sleqp_dyn_set_cons_weights_nogil(csleqp.SleqpFunc* func,
                                                           const double* cons_weights,
                                                           void* func_data) nogil:
  with gil:
    return sleqp_dyn_set_cons_weights(func,
                                      cons_weights,
                                      func_data)


cdef set_dyn_func_callbacks(csleqp.SleqpDynFuncCallbacks* callbacks):
  if release_gil:
    callbacks[0].set_value        = &sleqp_func_set_nogil
    callbacks[0].nonzeros         = &sleqp_func_nonzeros_nogil
    callbacks[0].set_error_bound  = &sleqp_dyn_set_error_bound_nogil
    callbacks[0].set_obj_weight   = &sleqp_dyn_set_obj_weight_nogil
    callbacks[0].set_cons_weights = &sleqp_dyn_set_cons_weights_nogil
    callbacks[0].eval             = &sleqp_dyn_eval_nogil
    callbacks[0].obj_grad         = &sleqp_func_obj_grad_nogil
    callbacks[0].cons_jac         = &sleqp_func_cons_jac_nogil
    callbacks[0].hess_prod        = &sleqp_func_hess_prod_nogil
  else:
    callbacks[0].set_value        = &sleqp_func_set
    callbacks[0].nonzeros         = &sleqp_func_nonzeros
    callbacks[0].set_error_bound  = &sleqp_dyn_set_error_bound
    callbacks[0].set_obj_weight   = &sleqp_dyn_set_obj_weight
    callbacks[0].set_cons_weights = &sleqp_dyn_set_cons_weights
    callbacks[0].eval             = &sleqp_dyn_eval
    callbacks[0].obj_grad         = &sleqp_func_obj_grad
    callbacks[0].cons_jac         = &sleqp_func_cons_jac
    callbacks[0].hess_prod        = &sleqp_func_hess_prod

  callbacks[0].func_free        = &sleqp_func_free


cdef object dyn_funcs = weakref.WeakSet()

cdef update_dyn_func_callbacks():
  cdef _Func func
  cdef csleqp.SleqpDynFuncCallbacks callbacks

  set_dyn_func_callbacks(&callbacks)

  for obj in dyn_funcs:
    func = <_Func> obj
    csleqp_call(csleqp.sleqp_dyn_func_set_callbacks(func.cfunc,
                                                    &callbacks))

cdef csleqp.SLEQP_RETCODE create_dyn_func(csleqp.SleqpFunc** cfunc,
                                          object func,
                                          int num_variables,
                                          int num_constraints):
  cdef csleqp.SleqpDynFuncCallbacks callbacks

  assert func is not None

  set_dyn_func_callbacks(&callbacks)

  return csleqp.sleqp_dyn_func_create(cfunc,
                                      &callbacks,
                                      num_variables,
                                      num_constraints,
                                      <void*> func)



#cython: language_level=3


cdef _readonly(arr):
    result = arr.view()
    result.flags.writeable = False
    return result


cdef csleqp.SLEQP_RETCODE create_problem(csleqp.SleqpProblem** problem,
                                         csleqp.SleqpFunc* cfunc,
                                         np.ndarray var_lb,
                                         np.ndarray var_ub,
                                         np.ndarray general_lb,
                                         np.ndarray general_ub,
                                         csleqp.SleqpSettings* csettings,
                                         object linear_coeffs = None,
                                         np.ndarray linear_lb = None,
                                         np.ndarray linear_ub = None):

  assert var_lb is not None
  assert var_ub is not None

  assert general_lb is not None
  assert general_ub is not None

  cdef csleqp.SleqpVec* var_lb_vec
  cdef csleqp.SleqpVec* var_ub_vec

  cdef csleqp.SleqpVec* general_lb_vec
  cdef csleqp.SleqpVec* general_ub_vec

  cdef csleqp.SleqpVec* linear_lb_vec
  cdef csleqp.SleqpVec* linear_ub_vec

  cdef csleqp.SleqpMat* linear_coeffs_mat

  cdef int num_vars = var_lb.shape[0]
  cdef int num_cons = general_lb.shape[0]

  assert cfunc != NULL

  cdef int num_linear_constraints = 0

  if linear_coeffs is not None:
    num_linear_constraints = linear_coeffs.shape[0]

  try:

    csleqp_call(csleqp.sleqp_vec_create_empty(&var_lb_vec,
                                              num_vars))

    csleqp_call(csleqp.sleqp_vec_create_empty(&var_ub_vec,
                                              num_vars))

    csleqp_call(csleqp.sleqp_vec_create_empty(&general_lb_vec,
                                              num_cons))

    csleqp_call(csleqp.sleqp_vec_create_empty(&general_ub_vec,
                                              num_cons))

    csleqp_call(csleqp.sleqp_mat_create(&linear_coeffs_mat,
                                        num_linear_constraints,
                                        num_vars,
                                        0))

    csleqp_call(csleqp.sleqp_vec_create_empty(&linear_lb_vec,
                                              num_linear_constraints))

    csleqp_call(csleqp.sleqp_vec_create_empty(&linear_ub_vec,
                                              num_linear_constraints))

    array_to_sleqp_sparse_vec(var_lb, var_lb_vec)
    array_to_sleqp_sparse_vec(var_ub, var_ub_vec)
    array_to_sleqp_sparse_vec(general_lb, general_lb_vec)
    array_to_sleqp_sparse_vec(general_ub, general_ub_vec)

    if linear_lb is not None:
      array_to_sleqp_sparse_vec(linear_lb, linear_lb_vec)

    if linear_ub is not None:
      array_to_sleqp_sparse_vec(linear_ub, linear_ub_vec)

    if linear_coeffs is not None:
      csleqp_call(matrix_to_sleqp_sparse_matrix(linear_coeffs,
                                                linear_coeffs_mat))

    csleqp_call(csleqp.sleqp_problem_create(problem,
                                            cfunc,
                                            var_lb_vec,
                                            var_ub_vec,
                                            general_lb_vec,
                                            general_ub_vec,
                                            linear_coeffs_mat,
                                            linear_lb_vec,
                                            linear_ub_vec,
                                            csettings))

    return csleqp.SLEQP_OKAY

  except BaseException as exception:
    return csleqp.SLEQP_ERROR

  finally:
    csleqp_call(csleqp.sleqp_vec_free(&linear_ub_vec))
    csleqp_call(csleqp.sleqp_vec_free(&linear_lb_vec))

    csleqp_call(csleqp.sleqp_mat_release(&linear_coeffs_mat))

    csleqp_call(csleqp.sleqp_vec_free(&general_ub_vec))
    csleqp_call(csleqp.sleqp_vec_free(&general_lb_vec))

    csleqp_call(csleqp.sleqp_vec_free(&var_ub_vec))
    csleqp_call(csleqp.sleqp_vec_free(&var_lb_vec))


cdef class _Problem:
  cdef csleqp.SleqpProblem* cproblem
  cdef _Func funcref

  def __cinit__(self):
    self.cproblem = NULL

  @staticmethod
  cdef _Problem create(csleqp.SleqpFunc* cfunc,
                       object var_lb,
                       object var_ub,
                       object cons_lb,
                       object cons_ub,
                       csleqp.SleqpSettings* csettings,
                       object linear_coeffs = None,
                       np.ndarray linear_lb = None,
                       np.ndarray linear_ub = None):

    cdef _Problem _problem = _Problem()

    var_lb = np.atleast_1d(var_lb)
    var_ub = np.atleast_1d(var_ub)

    cons_lb = np.atleast_1d(cons_lb)
    cons_ub = np.atleast_1d(cons_ub)

    cdef int num_vars = var_lb.shape[0]
    cdef int num_cons = cons_lb.shape[0]

    csleqp_call(create_problem(&_problem.cproblem,
                               cfunc,
                               var_lb,
                               var_ub,
                               cons_lb,
                               cons_ub,
                               csettings,
                               linear_coeffs,
                               linear_lb,
                               linear_ub))

    _problem.funcref = _Func()
    _problem.funcref.set_func(cfunc)

    return _problem

  @property
  def num_vars(self) -> int:
    return csleqp.sleqp_problem_num_vars(self.cproblem)

  @property
  def num_cons(self) -> int:
    return csleqp.sleqp_problem_num_cons(self.cproblem)

  @property
  def var_lb(self) -> np.array:
    return sleqp_sparse_vec_to_array(csleqp.sleqp_problem_vars_lb(self.cproblem))

  @property
  def var_ub(self) -> np.array:
    return sleqp_sparse_vec_to_array(csleqp.sleqp_problem_vars_ub(self.cproblem))

  @property
  def cons_lb(self) -> np.array:
    return sleqp_sparse_vec_to_array(csleqp.sleqp_problem_cons_lb(self.cproblem))

  @property
  def cons_ub(self) -> np.array:
    return sleqp_sparse_vec_to_array(csleqp.sleqp_problem_cons_ub(self.cproblem))

  @property
  def hess_struct(self) -> HessianStruct:
    return HessianStruct(self.funcref)

  def __dealloc__(self):
    csleqp_call(csleqp.sleqp_problem_release(&self.cproblem))


cdef class BaseProblem:
  """
  Base class modeling a nonlinear optimization problem (NLP)
  """
  cdef dict __dict__

  cdef object _problem
  cdef object _func
  cdef Settings _settings

  def __init__(self, problem):
    self._problem = problem

  @property
  def num_vars(self) -> int:
    """
    Number of variables in the problem
    """
    return self._problem.num_vars

  @property
  def num_cons(self) -> int:
    """
    Number of constraints in the problem
    """
    return self._problem.num_cons

  @property
  def var_lb(self) -> np.array:
    """
    Lower bounds on variables in [-oo, oo)
    """
    return _readonly(self._problem.var_lb)

  @property
  def var_ub(self) -> np.array:
    """
    Upper bounds on variables in (-oo, oo]
    """
    return _readonly(self._problem.var_ub)

  @property
  def cons_lb(self) -> np.array:
    """
    Lower bounds on constraints in [-oo, oo)
    """
    return _readonly(self._problem.cons_lb)

  @property
  def cons_ub(self) -> np.array:
    """
    Upper bounds on constraints in (-oo, oo]
    """
    return _readonly(self._problem.cons_ub)

  @property
  def func(self) -> object:
    """
    Underlying (user-provided) nonlinear function
    """
    return self._func

  @property
  def settings(self) -> object:
    return self._settings


cdef class Problem(BaseProblem):
  cdef _Func funcref

  def __cinit__(self,
                object func,
                object var_lb=[],
                object var_ub=[],
                object cons_lb=[],
                object cons_ub=[],
                Settings settings=None,
                **properties):

    cdef int num_vars = np.atleast_1d(var_lb).shape[0]
    cdef int num_cons = np.atleast_1d(cons_lb).shape[0]
    cdef csleqp.SleqpFunc* cfunc
    cdef csleqp.SleqpSettings* csettings

    if settings is None:
      self._settings = Settings()
    else:
      self._settings = settings

    csettings = <csleqp.SleqpSettings*> self._settings.settings

    self.funcref = _Func()

    csleqp_call(create_func(&cfunc,
                            func,
                            num_vars,
                            num_cons))

    assert cfunc != NULL

    self._func = func

    try:
      self.problem = _Problem.create(cfunc,
                                     var_lb,
                                     var_ub,
                                     cons_lb,
                                     cons_ub,
                                     csettings,
                                     properties.get('linear_coeffs', None),
                                     properties.get('linear_lb', None),
                                     properties.get('linear_ub', None))

      self.funcref.set_func(cfunc)
      funcs.add(self.funcref)

    finally:
      csleqp_call(csleqp.sleqp_func_release(&cfunc))

    try:
      func.set_hessian_struct(self.hess_struct)
    except AttributeError:
      pass

  # Actual arguments processed in __cinit__
  def __init__(self, *args, **kwds):
    super().__init__(self.problem)

  @property
  def hess_struct(self) -> HessianStruct:
    """
    Block structure of the Hessian matrix
    """
    return self.problem.hess_struct

  def _get_problem(self):
    return self.problem


cdef class LSQProblem(BaseProblem):
  """
  Class modeling a least-squares (LSQ) problem
  """

  cdef _Problem problem
  cdef _Func funcref

  def __cinit__(self,
                object func,
                num_residuals,
                object var_lb=[],
                object var_ub=[],
                object cons_lb=[],
                object cons_ub=[],
                Settings settings=None,
                **properties):

    cdef int num_vars = np.atleast_1d(var_lb).shape[0]
    cdef int num_cons = np.atleast_1d(cons_lb).shape[0]
    cdef csleqp.SleqpFunc* cfunc
    cdef csleqp.SleqpSettings* csettings

    if settings is None:
      self._settings = Settings()
    else:
      self._settings = settings

    csettings = <csleqp.SleqpSettings*> self._settings.settings

    self.funcref = _Func()

    csleqp_call(create_lsq_func(&cfunc,
                                func,
                                num_vars,
                                num_cons,
                                num_residuals,
                                properties.get('regularization', 0.),
                                csettings))

    assert cfunc != NULL

    self._func = func

    try:
      self.problem = _Problem.create(cfunc,
                                     var_lb,
                                     var_ub,
                                     cons_lb,
                                     cons_ub,
                                     csettings,
                                     properties.get('linear_coeffs', None),
                                     properties.get('linear_lb', None),
                                     properties.get('linear_ub', None))

      self.funcref.set_func(cfunc)
      lsq_funcs.add(self.funcref)

    finally:
      csleqp_call(csleqp.sleqp_func_release(&cfunc))

    try:
      func.set_hessian_struct(self.hess_struct)
    except AttributeError:
      pass

  # Actual arguments processed in __cinit__
  def __init__(self, *args, **kwds):
    super().__init__(self.problem)

  def _get_problem(self):
    return self.problem


cdef class DynProblem(BaseProblem):
  """
  Class modeling a dynamic problem, i.e.,
  a problem where evaluations have a limited,
  but controllable accuracy
  """

  cdef _Problem problem
  cdef _Func funcref

  def __cinit__(self,
                object func,
                object var_lb=[],
                object var_ub=[],
                object cons_lb=[],
                object cons_ub=[],
                Settings settings=None,
                **properties):

    cdef int num_vars = np.atleast_1d(var_lb).shape[0]
    cdef int num_cons = np.atleast_1d(cons_lb).shape[0]
    cdef csleqp.SleqpFunc* cfunc
    cdef csleqp.SleqpSettings* csettings

    if settings is None:
      self._settings = Settings()
    else:
      self._settings = settings

    csettings = <csleqp.SleqpSettings*> self._settings.settings

    self.funcref = _Func()

    csleqp_call(create_dyn_func(&cfunc,
                                func,
                                num_vars,
                                num_cons))

    assert cfunc != NULL
    self._func = func

    try:
      self.problem = _Problem.create(cfunc,
                                     var_lb,
                                     var_ub,
                                     cons_lb,
                                     cons_ub,
                                     csettings,
                                     properties.get('linear_coeffs', None),
                                     properties.get('linear_lb', None),
                                     properties.get('linear_ub', None))

      self.funcref.set_func(cfunc)
      dyn_funcs.add(self.funcref)

      pass
    finally:
      csleqp_call(csleqp.sleqp_func_release(&cfunc))
    try:
      func.set_hessian_struct(self.hess_struct)
    except AttributeError:
      pass

  # Actual arguments processed in __cinit__
  def __init__(self, *args, **kwds):
    super().__init__(self.problem)

  @property
  def hess_struct(self) -> HessianStruct:
    """
    Block structure of the Hessian matrix
    """
    return self.problem.hess_struct

  def _get_problem(self):
    return self.problem
